#!/usr/bin/env python

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2021 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

"""Validate A2L database according to ASAP2 rules (and more...).
"""

import enum
from collections import Counter, namedtuple
from itertools import combinations, filterfalse
from logging import getLogger

import pya2l.model as model
from pya2l import DB
from pya2l.api.inspect import AxisDescr, Characteristic, Measurement, ModCommon, ModPar


# *** Validator generated no diagnostic messages ***

Message = namedtuple("Message", "type category diag_code text")


class Level(enum.IntEnum):
    INFORMATION = 1
    WARNING = 2
    ERROR = 3


class Category(enum.IntEnum):
    DUPLICATE = 1
    MISSING = 2
    OBSOLETE = 3


class Diagnostics(enum.IntEnum):
    MULTIPLE_DEFINITIONS_IN_NAMESPACE = 1
    DEFINITION_IN_MULTIPLE_NAMESPACES = 2
    INVALID_C_IDENTIFIER = 3
    MISSING_BYTE_ORDER = 4
    MISSING_ALIGNMENT = 5
    MISSING_EPK = 6
    MISSING_ADDR_EPK = 7
    MISSING_MODULE = 8
    DEPRECATED = 9
    OVERLAPPING_MEMORY = 10


MAX_C_IDENTIFIER_LEN = 32  # ISO C90.


# any(len(e) > MAX_C_IDENTIFIER_LEN for e in 'CM.VTAB_RANGE.DEFAULT_VALUE.REF'.split("."))
# Some part of '{}' are longer than 32 chars (ISO C90 limit).


def names(objs):
    """ """
    return [o.name for o in objs]


class Validator:
    """
    Paramaters
    ----------
    session: Sqlite3 database object.
    """

    def __init__(self, session, loglevel="INFO"):
        self.logger = getLogger(self.__class__.__name__)
        # self.logger.setLevel("INFO")
        self._session = session
        self._diagnostics = []
        self._identifier = {}

    def __call__(self):
        """Run validation and return an iterable of diagnostics (Message)."""
        # reset previous diagnostics for repeated runs
        self._diagnostics = []
        self.check_top_level_structure()
        self._traverse_db()
        return tuple(self.diagnostics)

    def check_top_level_structure(self):
        self.modules = self.session.query(model.Module).all()
        if not self.modules:
            self.emit_diagnostic(
                Level.WARNING,
                Category.MISSING,
                Diagnostics.MISSING_MODULE,
                "A2l file requires at least one /MODULE.",
            )
        else:
            # Keep modules for subsequent checks
            pass

    @property
    def session(self):
        return self._session

    def _traverse_db(self):
        """ """

        """
        'a2ml', 'axis_pts', 'characteristic', 'compu_method',
        'compu_tab', 'compu_vtab', 'compu_vtab_range', 'frame', 'function', 'group', 'if_data', 'if_data_association',
        'longIdentifier', 'measurement', 'mod_common', 'mod_par', 'name', 'project', 'record_layout', 'unit',
        'user_rights', 'variant_coding'
        """
        for module in self.modules:
            self._validate_mod_common(module)
            self._validate_mod_par(module)

    def _validate_mod_common(self, module):
        mod_common = mod_common = ModCommon(self.session, module.name)
        if mod_common.byteOrder is None:
            self.emit_diagnostic(
                Level.WARNING,
                Category.MISSING,
                Diagnostics.MISSING_BYTE_ORDER,
                f"{module.name}::ModCommon: Missing BYTE_ORDER.",
            )
        missing_alignments = [e for e in mod_common.alignment.items() if e[1] is None]
        if missing_alignments:
            self.emit_diagnostic(
                Level.WARNING,
                Category.MISSING,
                Diagnostics.MISSING_ALIGNMENT,
                f"{module.name}::ModCommon: Missing ALIGNMENT(s): {[e[0] for e in missing_alignments]}.",
            )

    def _validate_mod_par(self, module):
        if not ModPar.exists(self.session, module.name):
            return
        mod_par = ModPar(self.session, module.name)
        if mod_par.epk is None:
            self.emit_diagnostic(
                Level.WARNING,
                Category.MISSING,
                Diagnostics.MISSING_EPK,
                f"{module.name}::ModPar: Missing EPK.",
            )
        elif not mod_par.adrEpk:
            self.emit_diagnostic(
                Level.WARNING,
                Category.MISSING,
                Diagnostics.MISSING_ADDR_EPK,
                f"{module.name}::ModPar: Missing ADDR_EPK.",
            )
        if mod_par.memoryLayouts:
            self.emit_diagnostic(
                Level.WARNING,
                Category.MISSING,
                Diagnostics.DEPRECATED,
                f"{module.name}::ModPar: MEMORY_LAYOUTs are deprecated, use MEMORY_SEGMENTs instead.",
            )
        # memorySegments

    '''
    def duplicate_ids(self, objs):
        """
        """
        cnt = Counter(names(objs))
        #cnt["CM.TAB_NOINTP.NO_DEFAULT_VALUE.REF"] = 3
        dups = filterfalse(lambda x: x[1] == 1, cnt.items())
        print(list(names(objs)), end = "\n\n")
        return list(dups)

    def _load_module_identifiers(self):
        TABLES = ("axis_pts", "characteristic", "compu_method", "compu_tab", "compu_vtab", "compu_vtab_range",
            "frame", "function", "group", "measurement", "mod_common", "mod_par", "record_layout", "unit",
            "user_rights", "variant_coding",
        )

    def check_uniqueness(self, module, tables):
        table_combinations = combinations(tables,2 )    # Pairwise.
        print("TC:", list(table_combinations))

    def check_namespaces(self):
        """
        """
        modules = self.session.query(model.Module).all()
        for module in modules:
            print(module)
            self.check_uniqueness(module, ("compu_vtab", "compu_vtab_range", "compu_tab"))
            self.check_uniqueness(module, ("axis_pts", "characteristic", "measurement"))
            dups = self.duplicate_ids(module.compu_tab)
            if dups:
                for dup, cnt in dups:
                    print("COMPU_TAB: multiple occurrences of identifier '{}'.".format(dup))
            print("TAB:", names(module.compu_tab), end = "\n\n")
            print("VTAB:", names(module.compu_vtab), end = "\n\n")
            print("VTAB-RANGE:", names(module.compu_vtab_range), end = "\n\n")
    '''

    def emit_diagnostic(self, level: Level, category: Category, diag: Diagnostics, message: str = None):
        try:
            self.logger.warning("%s - %s", level.name, message)
        except Exception:
            # Be robust even if logging formatting fails for any reason
            pass
        self._diagnostics.append(Message(level, category, diag, message))

    @property
    def diagnostics(self):
        return self._diagnostics


"""

class MODULE(Keyword):
    multiple = True
    block = True
    children = [
        "A2ML",
        "AXIS_PTS",
        "CHARACTERISTIC",
        "COMPU_METHOD",
        "COMPU_TAB",
        "COMPU_VTAB",
        "COMPU_VTAB_RANGE",
        "FRAME",
        "FUNCTION",
        "GROUP",
        "IF_DATA",
        "INSTANCE",
        "MEASUREMENT",
        "MOD_COMMON",
        "MOD_PAR",
        "RECORD_LAYOUT",
        "STRUCTURE_COMPONENT",
        "TYPEDEF_MEASUREMENT",
        "TYPEDEF_STRUCTURE",
        "UNIT",
        "USER_RIGHTS",
        "VARIANT_CODING",
    ]
"""



import atexit
import json
import os
import sys
from pprint import pprint

import pya2l.a2lparser_ext as ext
from pya2l import DB
from pya2l.aml.ifdata_parser import IfDataParser


os.chdir(r"C:\csprojects\pyA2L\pya2l\examples")

db = DB()
# session = db.open_create("XCPsim2")
session = db.open_create("ASAP2_Demo_V161")

ifdp = IfDataParser(session)
# pprint(ifdp.root)

IFD0 = """/begin IF_DATA XCP
/begin SEGMENT 0x01 0x02 0x00 0x00 0x00
/begin CHECKSUM XCP_ADD_44 MAX_BLOCK_SIZE 0xFFFF EXTERNAL_FUNCTION "" /end CHECKSUM
/begin PAGE 0x01 ECU_ACCESS_WITH_XCP_ONLY XCP_READ_ACCESS_WITH_ECU_ONLY XCP_WRITE_ACCESS_NOT_ALLOWED /end PAGE
/begin PAGE 0x00 ECU_ACCESS_WITH_XCP_ONLY XCP_READ_ACCESS_WITH_ECU_ONLY XCP_WRITE_ACCESS_WITH_ECU_ONLY /end PAGE
/end SEGMENT
/end IF_DATA"""

pprint(ifdp.parse(IFD0))

IFD1 = """/begin IF_DATA XCP
      /begin PROTOCOL_LAYER
        0x0100
        0x07D0
        0x2710
        0x00
        0x00
        0x00
        0x00
        0x00
        0xA6
        0x28
        BYTE_ORDER_MSB_LAST
        ADDRESS_GRANULARITY_BYTE
        OPTIONAL_CMD GET_COMM_MODE_INFO
        OPTIONAL_CMD GET_ID
        OPTIONAL_CMD SET_MTA
        OPTIONAL_CMD UPLOAD
        OPTIONAL_CMD SHORT_UPLOAD
        OPTIONAL_CMD BUILD_CHECKSUM
        OPTIONAL_CMD TRANSPORT_LAYER_CMD
        OPTIONAL_CMD USER_CMD
        OPTIONAL_CMD DOWNLOAD
        OPTIONAL_CMD DOWNLOAD_NEXT
        OPTIONAL_CMD SHORT_DOWNLOAD
        OPTIONAL_CMD MODIFY_BITS
        OPTIONAL_CMD SET_CAL_PAGE
        OPTIONAL_CMD GET_CAL_PAGE
        OPTIONAL_CMD GET_PAG_PROCESSOR_INFO
        OPTIONAL_CMD GET_PAGE_INFO
        OPTIONAL_CMD SET_DAQ_PTR
        OPTIONAL_CMD WRITE_DAQ
        OPTIONAL_CMD SET_DAQ_LIST_MODE
        OPTIONAL_CMD GET_DAQ_LIST_MODE
        OPTIONAL_CMD START_STOP_DAQ_LIST
        OPTIONAL_CMD START_STOP_SYNCH
        OPTIONAL_CMD GET_DAQ_CLOCK
        OPTIONAL_CMD GET_DAQ_PROCESSOR_INFO
        OPTIONAL_CMD GET_DAQ_RESOLUTION_INFO
        OPTIONAL_CMD GET_DAQ_EVENT_INFO
        OPTIONAL_CMD FREE_DAQ
        OPTIONAL_CMD ALLOC_DAQ
        OPTIONAL_CMD ALLOC_ODT
        OPTIONAL_CMD ALLOC_ODT_ENTRY
        OPTIONAL_CMD PROGRAM_START
        OPTIONAL_CMD PROGRAM_CLEAR
        OPTIONAL_CMD PROGRAM
        OPTIONAL_CMD PROGRAM_RESET
        OPTIONAL_CMD PROGRAM_PREPARE
        COMMUNICATION_MODE_SUPPORTED BLOCK
          SLAVE
          MASTER
            0x2B
            0x00
      /end PROTOCOL_LAYER
      /begin PAG
        0x01
      /end PAG
      /begin PGM
        PGM_MODE_ABSOLUTE
        0x01
        0x08
      /end PGM
      /begin DAQ
        DYNAMIC
        0x00
        0x05
        0x00
        OPTIMISATION_TYPE_DEFAULT
        ADDRESS_EXTENSION_FREE
        IDENTIFICATION_FIELD_TYPE_ABSOLUTE
        GRANULARITY_ODT_ENTRY_SIZE_DAQ_BYTE
        0x07
        OVERLOAD_INDICATION_PID
        PRESCALER_SUPPORTED
        /begin TIMESTAMP_SUPPORTED
          0x03
          SIZE_WORD
          UNIT_1MS
        /end TIMESTAMP_SUPPORTED
        /begin EVENT
          "5ms_Task"
          "5ms_Task"
          0x00
          DAQ
          0x01
          0x05
          0x06
          0x00
        /end EVENT
        /begin EVENT
          "10ms_Task"
          "10ms_Tas"
          0x01
          DAQ
          0x01
          0x0A
          0x06
          0x00
        /end EVENT
        /begin EVENT
          "20ms_Task"
          "20ms_Tas"
          0x02
          DAQ
          0x01
          0x14
          0x06
          0x00
        /end EVENT
        /begin EVENT
          "40ms_Task"
          "40ms_Tas"
          0x03
          DAQ
          0x01
          0x28
          0x06
          0x00
        /end EVENT
        /begin EVENT
          "80ms_Task"
          "80ms_Tas"
          0x04
          DAQ
          0x01
          0x50
          0x06
          0x00
        /end EVENT
      /end DAQ
      /begin XCP_ON_CAN
        0x0100
        CAN_ID_MASTER 0x01E2
        CAN_ID_SLAVE 0x9BFC1010
        BAUDRATE 0x07A120
        SAMPLE_POINT 0x4B
        SAMPLE_RATE SINGLE
        BTL_CYCLES 0x08
        SJW 0x02
        SYNC_EDGE SINGLE
        MAX_DLC_REQUIRED
      /end XCP_ON_CAN
    /end IF_DATA
    /begin IF_DATA CANAPE_MODULE
    /end IF_DATA"""

pprint(ifdp.parse(IFD1))



#!/usr/bin/env python

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2025 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

"""Functions adding semantics to model elements.
"""

import bisect
import math
import re
import typing
from dataclasses import dataclass
from operator import itemgetter

import numpy as np
from scipy import interpolate
from scipy.interpolate import RegularGridInterpolator

from pya2l import exceptions


try:
    import numexpr
except ImportError:
    has_numexpr = False
else:
    has_numexpr = True

POW = re.compile(r"pow\s*\((?P<params>.*?)\s*\)", re.IGNORECASE)
SYSC = re.compile(r"sysc\s*\((?P<param>.*?)\s*\)", re.IGNORECASE)


@dataclass
class Coeffs:
    a: float
    b: float
    c: float
    d: float
    e: float
    f: float


@dataclass
class CoeffsLinear:
    a: float
    b: float


def fix_axis_par(offset: int, shift: int, num_apo: int) -> list:
    """"""  # noqa: DAR101, DAR201
    return np.array([offset + (i * (2**shift)) for i in range(num_apo)], dtype="float64")


def fix_axis_par_dist(offset: int, distance: int, num_apo: int) -> list:
    """"""  # noqa: DAR101, DAR201
    return np.array([offset + (i * distance) for i in range(num_apo)], dtype="float64")


class Interpolate1D:
    """1-D linear interpolation.

    Parameters
    ----------

    pairs: iterable of 2-tuples (x,y)
        x: float
            X values
        y: float
            Y values -- must be of equal length than xs.
    saturate: bool
        - ``True``
            - if X value is less then or equal to min(xs), the result is set to the Y value
              corresponding to the lowest X value
            - if X value is greater than or equal to the highest X value, the result is set to the Y value
              corresponding to the highest X value.
    default: float or None
        returned if x is out of boundaries.
    """

    def __init__(self, pairs, saturate=True):
        xs, ys = zip(*pairs)

        if any(x1 - x0 <= 0 for x0, x1 in zip(xs, xs[1:])):
            raise ValueError("'xs' must be in strictly increasing order.")
        self.min_x = min(xs)
        self.max_x = max(xs)

        if saturate:
            fill_value = (ys[0], ys[-1])
            bounds_error = False
        else:
            fill_value = None
            bounds_error = True

        self.interp = interpolate.interp1d(x=xs, y=ys, kind="linear", bounds_error=bounds_error, fill_value=fill_value)

    def __call__(self, x):
        """Interpolate a single value.

        Parameters
        ----------
        x: float
        """
        return self.interp(x)


'''
class Lookup:
    """Table lookup

    Parameters
    ----------

    """

    def __init__(self, xs, ys, saturate = False, y_low = None, y_high = None):
        if any(x1 -x0 <= 0 for x0, x1 in zip(xs, xs[1 : ])):
            raise ValueError("'xs' must be in strictly increasing order.")
        self.xs = xs
        self.ys = ys
        self.saturate = saturate
        self.y_low = y_low
        self.y_high = y_high

    def __call__(self, x):
        """
        """
        if self.saturate:
            if x <= self.min_x:
                return self.y_low or self.ys[0]
            elif x >= self.max_x:
                return self.y_high or self.ys[-1]
        else:
            if not (self.xs[0] <= x <= self.xs[-1]):
                raise ValueError("x out of bounds")
        pos = bisect.bisect_right(self.xs, x) - 1
        return self.ys[pos]
'''


def axis_rescale(no_rescale_x: int, no_axis_pts: int, axis, virtual):
    """
    Parameters
    ----------
    no_rescale_x: int

    no_axis_pts: int
        Number of axis-points to calculate; shall be a power of two plus one, e.g. 5, 17, 33...

    axis: list-like

    virtual: list-like


    Returns
    -------
    `numpy.array`
    """
    k = 1
    d = (virtual[-1] - virtual[0] + 1) // (no_axis_pts - 1)
    xs = [axis[0]]
    for idx in range(no_rescale_x - 1):
        while True:
            kdv = (k * d) + virtual[0]
            if kdv >= virtual[idx + 1]:
                break
            k += 1
            x = axis[idx] + (((k - 1) * d) - virtual[idx]) * (axis[idx + 1] - axis[idx]) / (virtual[idx + 1] - virtual[idx])
            xs.append(x)
    xs.append(axis[no_rescale_x - 1])
    return np.array(xs)


class NormalizationAxes:
    """
    Appendix C: Using Reference Curves as Normalization Axes for Maps

    Parameters
    ----------
    x_norm:

    y_norm:

    z_map
    """

    def __init__(self, x_norm, y_norm, z_map):
        self.xn = np.array(x_norm)
        self.yn = np.array(y_norm)
        self.zm = np.array(z_map)
        self.ip_x = Interpolate1D(pairs=zip(self.xn[:, 0], self.xn[:, 1]))
        self.ip_y = Interpolate1D(pairs=zip(self.yn[:, 0], self.yn[:, 1]))
        x_size, y_size = self.zm.shape
        self.ip_m = RegularGridInterpolator((np.arange(x_size), np.arange(y_size)), self.zm, method="linear")

    def __call__(self, x, y):
        """
        Parameters
        ----------
        x: float

        y: float

        Returns
        -------
        float
        """
        x_new = self.ip_x(x)
        y_new = self.ip_y(y)
        return self.ip_m((y_new, x_new))


class RatFunc:
    """Evaluate rational function.

    Parameters
    ----------
    coeffs: Coeffs
        a, b, c, d, e, f - coefficients for the specified formula:
        f(x) = (axx + bx + c) / (dxx + ex + f)

        INT = f(PHYS)

        When using :method:`int_to_physical`: Restrictions have to be defined
        because this general equation cannot always be inverted.

    Note
    ----
        `Linear` is PHYS = f(INT)!!!
    """

    def __init__(self, coeffs: Coeffs):
        a, b, c, d, e, f = (coeffs.a, coeffs.b, coeffs.c, coeffs.d, coeffs.e, coeffs.f)
        self.p = np.poly1d([a, b, c])
        self.q = np.poly1d([d, e, f])
        if self.p.order == 1 and self.q.order == 0:
            self.p_inv = np.poly1d([(f / b), -((f * c) / (b * f))])
        else:
            self.p_inv = None
        self.coeffs = coeffs

    def physical_to_int(self, p):
        """Evaluate function PHYS ==> INT

        Parameters
        ----------
        p: int or float, scalar or numpy.ndarray
        """
        return self.p(p) / self.q(p)

    def int_to_physical(self, value):
        """Evaluate function INT ==> PHYS

        Parameters
        ----------
        i: int or float, scalar or numpy.ndarray

        Note
        ----
        Currently inversion of quadratic functions isn't supported, only linear ones.
        """
        if self.p.order == 1 and self.q.order == 0:
            return self.p_inv(value)
        elif self.p.order == 0 and self.q.order == 1:
            return self.physical_to_int(value)  # Inverse functions are already inverted...
        elif self.p.order == 0 and self.q.order == 0:
            raise exceptions.MathError("Cannot invert constant function.")
        else:
            raise NotImplementedError("Cannot invert quadratic function.")

    def __str__(self) -> str:
        return f"RatFunc(coeffs=(a={self.coeffs.a}, b={self.coeffs.b}, c={self.coeffs.c}, d={self.coeffs.d}, e={self.coeffs.e}, f={self.coeffs.f}))"

    __repr__ = __str__


class Identical:
    """Identity function."""

    def __init__(self):
        pass

    def int_to_physical(self, i):
        if isinstance(i, np.ndarray):
            return i.copy()
        return i

    def physical_to_int(self, p):
        return p

    def __str__(self) -> str:
        return "Identical()"

    __repr__ = __str__


class Linear:
    """Evaluate linear function.

    Parameters
    ----------
    coeffs: CoeffsLinear
        a, b - coefficients for the specified formula:
        coefficients for the specified formula:
        f(x) = ax + b

        PHYS = f(INT)

    Note
    ----
        `RatFunc` is INT = f(PHYS)!!!
    """

    def __init__(self, coeffs: CoeffsLinear):
        a, b = coeffs.a, coeffs.b
        self.p = np.poly1d([a, b])
        self.a = a
        self.b = b

    def int_to_physical(self, i):
        """"""  # noqa: DAR101, DAR201
        return self.p(i)

    def _eval_pti(self, p):
        return (self.p - p).roots[0]

    def physical_to_int(self, p):
        """"""  # noqa: DAR101, DAR201
        if hasattr(p, "__iter__"):
            return np.array([self._eval_pti(i) for i in p])
        else:
            return self._eval_pti(p)

    def __str__(self) -> str:
        return f"Linear(coeffs=(a={self.a}, b={self.b}))"

    __repr__ = __str__


class LookupTable:
    """Basic lookup table.
    An integer value is mapped to an integer or display string.

    Parameters
    ----------
        mapping: iterable of 2-tuples (key, value)
            - keys can be either floats or ints (internaly converted to int)
            - values are either integers or strings.

        default: int or str
            returned if value is not in mapping.
    """

    def __init__(self, mapping, default=None):
        mapping = ((int(item[0]), item[1]) for item in mapping)
        self.mapping = dict(mapping)
        self.mapping_inv = {v: k for k, v in self.mapping.items()}
        self.default = default
        self.map_internal_to_phys_vec = np.vectorize(self.map_internal_to_phys)

    def map_internal_to_phys(self, value):
        return self.mapping.get(value, self.default)

    def int_to_physical(self, value):
        """"""  # noqa: DAR101, DAR201
        if hasattr(value, "__iter__"):
            return self.map_internal_to_phys_vec(value)
        else:
            return self.map_internal_to_phys(value)

    def physical_to_int(self, p):
        """"""  # noqa: DAR101, DAR201
        if hasattr(p, "__iter__") and not isinstance(p, str):
            return [self.mapping_inv.get(r) for r in p]
        else:
            return self.mapping_inv.get(p)


class InterpolatedTable:
    """Table with linear interpolation."""

    def __init__(self, pairs, default=None):
        self.interp = Interpolate1D(pairs, saturate=False)
        self.default = default

    def int_to_physical(self, i):
        """"""  # noqa: DAR101, DAR201
        try:
            return self.interp(i)
        except ValueError:
            return self.default

    def physical_to_int(self, p):
        """"""  # noqa: DAR101, DAR201, DAR401
        raise NotImplementedError()

    def __str__(self) -> str:
        return f"InterpolatedTable(x={self.interp.interp.x}, y={self.interp.interp.y}, max_x={self.interp.max_x}, max_y={self.interp.max_y})"

    __repr__ = __str__


class LookupTableWithRanges:
    """Lookup table where keys define numerical ranges.

    A value range is mapped to a display string.

    Parameters
    ----------
        mapping: iterable of 3-tuples (key_min, key_max, value)
            - keys can be either floats or ints (s. `dtype`).
            - values are strings.

        default: str
            returned if value is out of any declared range.

        dtype: int | float
            Datatype of keys.
    """

    def __init__(self, mapping, default=None, dtype=int):
        if not (dtype is int or dtype is float):
            raise ValueError("dtype must be either int or float")
        mapping = ((dtype(item[0]), dtype(item[1]), item[2]) for item in mapping)
        self.mapping = sorted(mapping, key=itemgetter(0))
        self.min_values = [item[0] for item in self.mapping]
        self.max_values = [item[1] for item in self.mapping]
        self.minimum = min(self.min_values)
        self.maximum = max(self.max_values)
        self.display_values = [item[2] for item in self.mapping]
        self.dict_inv = dict(zip(self.display_values, self.min_values))  # min_value, according to spec.
        self.default = default
        if dtype == int:
            self.in_range = lambda x, left, right: left <= x <= right
        else:
            self.in_range = lambda x, left, right: left <= x < right

    def _lookup(self, x):
        """"""  # noqa: DAR101, DAR201
        if not (self.minimum <= x <= self.maximum):
            return self.default
        pos = bisect.bisect_right(self.min_values, x) - 1
        display_value = self.display_values[pos]
        min_value = self.min_values[pos]
        max_value = self.max_values[pos]
        if self.in_range(x, min_value, max_value):
            return display_value
        else:
            return self.default

    def int_to_physical(self, i):
        """"""  # noqa: DAR101, DAR201
        if hasattr(i, "__iter__"):
            return np.array([self._lookup(r) for r in i])
        else:
            return self._lookup(i)

    def physical_to_int(self, p):
        """"""  # noqa: DAR101, DAR201
        if hasattr(p, "__iter__") and not isinstance(p, str):
            return [self.dict_inv.get(r, None) for r in p]
        else:
            return self.dict_inv.get(p, None)


class FormulaBase:
    """Base class for formula interpreters.

    Parameters
    ----------

    formula: str
        function to calculate the physical value from the control unit internal value.

    inverse_formula: str
        function for calculation of the control unit internal value from the physical value.

    system_constants: list of 2-tuples (name, value)
    """

    def __init__(
        self, formula: str, inverse_formula: typing.Optional[str] = None, system_constants: typing.Optional[typing.Dict] = None
    ):
        if not formula:
            raise ValueError("Formula cannot be None or empty.")
        if system_constants:
            self.system_constants = system_constants
        else:
            self.system_constants = {}
        self.formula = self._replace_special_symbols(formula)
        self.inverse_formula = self._replace_special_symbols(inverse_formula) if inverse_formula else None

    def sysc(self, key):
        return self.system_constants[key]

    def _build_namespace(self, *args):
        if len(args) == 0:
            raise ValueError("Formula called with no paramters.")
        xs = {f"X{i}": v for i, v in enumerate(args, 1)}
        if len(args) == 1:  # In this case...
            xs["X"] = xs.get("X1")  # ... create an alias.
        namespace = self.MATH_FUNCS
        namespace.update(xs)
        namespace.update(self.system_constants)
        namespace["sysc"] = self.sysc
        return namespace

    @staticmethod
    def names_tolower(text):
        MATH_FUNCTIONS = re.compile("abs|acos|asin|atan|cos|cosh|exp|log|log10|pow|sin|sinh|sqrt|tan|tanh", re.IGNORECASE)
        return MATH_FUNCTIONS.sub(lambda m: m.group(0).lower(), text)


if has_numexpr:

    class Formula(FormulaBase):
        """ASAP2 formula interpreter based on *numexpr*.

        Parameters
        ----------

        formula: str
            function to calculate the physical value from the control unit internal value.

        inverse_formula: str
            function for calculation of the control unit internal value from the physical value.

        system_constants: list of 2-tuples (name, value)
        """

        MATH_FUNCS = {}

        def _replace_special_symbols(self, text):
            result = (
                text.replace("&&", " and ")
                .replace("||", " or ")
                .replace("!", "not ")
                .replace("acos", "arccos")
                .replace("asin", "arcsin")
                .replace("atan", "arctan")
            )
            text = FormulaBase.names_tolower(result)
            while True:
                # replace 'pow(a, b)' with '(a ** b)'
                match = POW.search(result)
                if match:
                    params = [p.strip() for p in match.group("params").split(",")]
                    assert len(params) == 2
                    pow_expr = "({} ** {})".format(*params)
                    head = result[: match.start()]
                    tail = result[match.end() :]
                    result = f"{head}{tail}{pow_expr}"
                else:
                    break
            while True:
                match = SYSC.search(result)  # replace 'sysc(a)' with value of 'a'
                if match:
                    param = match.group("param").strip()
                    value = self.sysc(param)
                    head = result[: match.start()]
                    tail = result[match.end() :]
                    result = f"{head}{tail}{value}"
                else:
                    break
            return result

        def int_to_physical(self, *args):
            """"""  # noqa: DAR101, DAR201
            try:
                return numexpr.evaluate(self.formula, local_dict=self._build_namespace(*args))
            except Exception as e:
                print(f"Error evaluating formula: {e!r})")
                return np.array([])

        def physical_to_int(self, *args):
            """"""  # noqa: DAR101, DAR201, DAR401
            if self.inverse_formula is None:
                raise NotImplementedError("Formula: physical_to_int() requires inverse_formula.")
            try:
                return numexpr.evaluate(self.inverse_formula, local_dict=self._build_namespace(*args))
            except Exception as e:
                print(f"Error evaluating inverse formula: {e!r})")
                return np.array([])

else:

    class Formula(FormulaBase):
        """Crude ASAP2 formula interpreter using python `eval()`.

        Parameters
        ----------

        formula: str
            function to calculate the physical value from the control unit internal value.

        inverse_formula: str
            function for calculation of the control unit internal value from the physical value.

        system_constants: list of 2-tuples (name, value)
        """

        MATH_FUNCS: typing.ClassVar = {
            "abs": math.fabs,
            "acos": math.acos,
            "asin": math.asin,
            "atan": math.atan,
            "cos": math.cos,
            "cosh": math.cosh,
            "exp": math.exp,
            "log": math.log,
            "log10": math.log10,
            "pow": math.pow,
            "sin": math.sin,
            "sinh": math.sinh,
            "sqrt": math.sqrt,
            "tan": math.tan,
            "tanh": math.tanh,
        }

        def _replace_special_symbols(self, text):
            text = FormulaBase.names_tolower(text)
            return text.replace("&&", " and ").replace("||", " or ").replace("!", "not ")

        def int_to_physical(self, *args):
            """"""  # noqa: DAR101, DAR201
            return eval(self.formula, dict(), self._build_namespace(*args))

        def physical_to_int(self, *args):
            """"""  # noqa: DAR101, DAR201, DAR401
            if self.inverse_formula is None:
                raise NotImplementedError("Formula: physical_to_int() requires inverse_formula.")
            return eval(self.inverse_formula, dict(), self._build_namespace(*args))



__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2022-2025 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import argparse
import sys
from pathlib import Path

from pya2l import DB
from pya2l.version import __version__


def main():
    parser = argparse.ArgumentParser(description="Import/export from/to a2l(db) files.")

    group = parser.add_argument_group()
    group.add_argument(
        "-o",
        "--output",
        help="output file when exporting",
        dest="output",
        type=str,
        metavar="output_file",
    )

    parser.add_argument(
        "-E",
        "--encoding",
        help="Import file encoding, like ascii, latin-1, utf-8, ...",
        dest="encoding",
        type=str,
        default=None,
    )
    parser.add_argument(
        "-L",
        "--local-directory",
        help="If option is used, create DB in current working directory, else use import path.",
        action="store_true",
        default=False,
        dest="local",
    )
    parser.add_argument(
        "-l",
        help="Verbosity of log messages",
        choices=["warn", "info", "error", "debug", "critical", "WARN", "INFO", "ERROR", "DEBUG", "CRITICAL"],
        dest="loglevel",
        type=str,
        default="info",
    )

    parser.add_argument(
        "-p",
        help="Disable progress bar",
        action="store_true",
        dest="no_progress_bar",
    )

    parser.add_argument(
        "-V",
        help="Print pya2ldb version information and exit.",
        action="store_true",
        dest="version",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-i",
        "--import",
        help="A2L file to import",
        dest="ifn",
        type=str,
        metavar="A2L_file",
    )
    group.add_argument(
        "-e",
        "--export",
        help="A2LDB file to export",
        dest="efn",
        type=str,
        metavar="DB_file",
    )

    # parser.add_argument("-f", "--force-overwrite", help = "Force overwrite of existing file",
    # default = False, action = "store_true")

    args = parser.parse_args()
    if args.version:
        print(f"pya2ldb version: {__version__}")
        sys.exit(1)
    if not (args.ifn or args.efn):
        print("Either -i or -e option is required.")
        sys.exit(2)
    db = DB()

    if args.efn:
        efn = Path(args.efn)
        db.export_a2l(efn, args.output or sys.stdout)
    else:
        ifn = Path(args.ifn)
        session = db.import_a2l(
            ifn, encoding=args.encoding, loglevel=args.loglevel, local=args.local, progress_bar=not args.no_progress_bar
        )
        session.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2019 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

__author__ = "Christoph Schueler"
__version__ = "0.9"

##
## Convenience functions for Mako Templates.
##

from io import StringIO

from mako import exceptions
from mako.runtime import Context
from mako.template import Template

# from csstuff import strings


def indentText(text: str, leftmargin: int = 0):
    return "\n".join(
        [
            "%s%s"
            % (
                (" " * leftmargin),
                line,
            )
            if line
            else ""
            for line in text.splitlines()
        ]
    )


# TODO: rename to 'renderTemplate'!


def doTemplate(
    tmpl,
    namespace={},
    leftMargin=0,
    rightMargin=80,
    formatExceptions=True,
    encoding="utf-8",
):
    buf = StringIO()
    ctx = Context(buf, **namespace)
    try:
        tobj = Template(
            filename=tmpl, output_encoding=encoding, format_exceptions=formatExceptions
        )  # , imports ='re' # TODO: imports parameter.
        tobj.render_context(ctx)
    except Exception:
        print(exceptions.text_error_template().render())
        return None
    # return strings.reformat(buf.getvalue(), leftMargin, rightMargin)
    return buf.getvalue()


def doTemplateFromText(
    tmpl,
    namespace={},
    leftMargin=0,
    rightMargin=80,
    formatExceptions=True,
    encoding="utf-8",
):
    buf = StringIO()
    ctx = Context(buf, **namespace)
    try:
        tobj = Template(text=tmpl, output_encoding=encoding, format_exceptions=formatExceptions)  # , imports ='re'
        tobj.render_context(ctx)
    except Exception:
        print(exceptions.text_error_template().render())
        return None
    return indentText(buf.getvalue(), leftMargin)  # , rightMargin)


def callDef(template, definition, *args, **kwargs):
    return template.get_def(definition).render(*args, **kwargs)



import unittest

from pya2l.utils import cygpathToWin


class TestCygpath(unittest.TestCase):
    def test01(self):
        self.assertEquals(
            cygpathToWin("/cygdrive/c/projects/foobar/flonz"),
            r"c:\projects\foobar\flonz",
        )

    def test02(self):
        self.assertEquals(cygpathToWin(r"c:\projects\foobar\flonz"), r"c:\projects\foobar\flonz")


def main():
    unittest.main()


if __name__ == "__main__":
    main()



"""
IF_DATA Parser Module

This module provides functionality for parsing IF_DATA sections in A2L files.
It uses a token-based approach with a syntax tree to interpret the structure
of IF_DATA sections according to AML definitions.

The main class is IfDataParser which handles the parsing of IF_DATA sections.
"""

from collections import defaultdict
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Any, Dict, List, Optional, Union

import pya2l.model as model
from pya2l.a2lparser_ext import AmlType, unmarshal
from pya2l.aml.ifdata_lexer import IfDataLexer
from pya2l.logger import Logger


class IfDataTokenType(IntEnum):
    """Token types used by the IF_DATA parser."""

    NONE = 0
    IDENT = 1
    FLOAT = 2
    INT = 3
    COMMENT = 4
    STRING = 6
    BEGIN = 7
    END = 8
    WS = 9


@dataclass
class IfDataToken:
    """Represents a token in the IF_DATA section."""

    type: IfDataTokenType
    value: Any


class AMLPredefinedTypeEnum(IntEnum):
    """Predefined types in AML."""

    CHAR = 0
    INT = 1
    LONG = 2
    UCHAR = 3
    UINT = 4
    ULONG = 5
    INT64 = 6
    UINT64 = 7
    DOUBLE = 8
    FLOAT = 9
    FLOAT16 = 10


class ReferrerType(IntEnum):
    """Types of referrers in AML."""

    Enumeration = 0
    StructType = 1
    TaggedStructType = 2
    TaggedUnionType = 3


@dataclass
class Root:
    """Root node of the AML tree."""

    members: List[Any] = field(default_factory=list)


@dataclass
class Referrer:
    """Reference to another node in the AML tree."""

    category: ReferrerType
    identifier: str


@dataclass
class Struct:
    """Represents a struct in AML."""

    name: Optional[str] = field(default=None)
    members: List[Any] = field(default_factory=list)


@dataclass
class Member:
    """Represents a member of a struct in AML."""

    node: Any
    is_block: bool


@dataclass
class TaggedStruct:
    """Represents a tagged struct in AML."""

    name: Optional[str] = field(default=None)
    members: Dict[str, Any] = field(default_factory=dict)


@dataclass
class TaggedStructMember:
    """Represents a member of a tagged struct in AML."""

    definition: Any
    multiple: bool


@dataclass
class TaggedStructDefinition:
    """Represents the definition of a tagged struct in AML."""

    member: Any
    multiple: bool


@dataclass
class TaggedUnion:
    """Represents a tagged union in AML."""

    name: Optional[str] = field(default=None)
    members: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Enumeration:
    """Represents an enumeration in AML."""

    name: Optional[str] = field(default=None)
    values: Dict[str, Any] = field(default_factory=dict)


@dataclass
class PDT:
    """Represents a predefined type in AML."""

    type: AMLPredefinedTypeEnum
    arr_spec: List[int]


@dataclass
class Block:
    """Represents a block in AML."""

    tag: Optional[str]
    type: Any


@dataclass
class NullObject:
    """Represents a null object in AML."""

    pass


class EOFReached(Exception):
    """Signals end of token stream."""

    pass


class ParsingError(Exception):
    """Signals an error during parsing."""

    pass


def create_ref_dict(tree) -> defaultdict:
    """
    Create a dictionary of references from the AML tree.

    Args:
        tree: The AML tree root

    Returns:
        A dictionary mapping referrer types to dictionaries of named elements
    """
    result: defaultdict = defaultdict(dict)
    members = tree.members
    non_blocks = [m for m in members if not isinstance(m, (Block, NullObject))]

    for member in non_blocks:
        if isinstance(member, Struct):
            tp = ReferrerType.StructType
        elif isinstance(member, TaggedStruct):
            tp = ReferrerType.TaggedStructType
        elif isinstance(member, TaggedUnion):
            tp = ReferrerType.TaggedUnionType
        elif isinstance(member, Enumeration):
            tp = ReferrerType.Enumeration
        else:
            raise TypeError(f"Unsupported member type: {type(member).__name__}")

        result[tp][member.name] = member

    return result


def toplevel_ifdata(tree) -> Any:
    """
    Find the IF_DATA block in the AML tree.

    Args:
        tree: The AML tree root

    Returns:
        The IF_DATA block type

    Raises:
        ValueError: If the tree structure is not as expected
        TypeError: If no blocks are found
    """
    members = [m for m in tree.members if not isinstance(m, NullObject)]
    blocks = [m for m in members if isinstance(m, Block)]

    if blocks:
        for member in blocks:
            if isinstance(member, Block) and member.tag == "IF_DATA":
                return member.type
        raise ValueError("No IF_DATA block found in the AML tree")
    elif len(members) == 1:
        return members[0]
    else:
        raise ValueError("Invalid AML tree structure")


class IfDataParser:
    """
    Parser for IF_DATA sections in A2L files.

    This class parses IF_DATA sections according to AML definitions.
    It uses a token-based approach with a syntax tree to interpret
    the structure of IF_DATA sections.
    """

    def __init__(self, session, loglevel: str = "INFO") -> None:
        """
        Initialize the IF_DATA parser.

        Args:
            session: The database session containing AML definitions
            loglevel: The logging level
        """
        self.logger = Logger("IF_DATA", loglevel)
        self.root = None
        self.syntax_stack = []
        self.ref_dict = None
        self.tokens = []
        self.token_idx = 0
        self.level = 0
        self.num_tokens = 0

        # Initialize from session
        aml_section = session.query(model.AMLSection).first()
        if aml_section:
            aml_root = unmarshal(aml_section.parsed)
            self.root = self.traverse(aml_root)
            if self.root and self.root.members:
                try:
                    self.syntax_stack = [toplevel_ifdata(self.root)]
                    self.ref_dict = create_ref_dict(self.root)
                except (ValueError, TypeError) as e:
                    self.logger.error(f"Failed to initialize parser: {str(e)}")
                    self.root = None

    def parse(self, data) -> Dict:
        """
        Parse an IF_DATA section.

        Args:
            data: The IF_DATA section text

        Returns:
            A dictionary representing the parsed IF_DATA section

        Raises:
            ParsingError: If parsing fails
        """
        if self.root is None:
            self.logger.warn("Parser not initialized properly, returning empty result")
            return {}

        try:
            # Tokenize the input
            lexer = IfDataLexer(data)
            self.tokens = lexer.run()
            self.token_idx = 0
            self.level = 0
            self.num_tokens = len(self.tokens)

            # Parse the IF_DATA section
            self.match(IfDataTokenType.BEGIN)
            self.match(IfDataTokenType.IDENT, "IF_DATA")
            result = self.enter(self.syntax_tos)
            self.leave()
            self.match(IfDataTokenType.END)
            self.match(IfDataTokenType.IDENT, "IF_DATA")

            return result
        except EOFReached:
            self.logger.error("Unexpected end of input")
            raise ParsingError("Unexpected end of input")
        except Exception as e:
            self.logger.error(f"Error parsing IF_DATA: {str(e)}")
            raise ParsingError(f"Error parsing IF_DATA: {str(e)}")

    def block(self) -> Dict:
        """
        Parse a block in the IF_DATA section.

        Returns:
            A dictionary representing the parsed block

        Raises:
            TypeError: If the token type is not as expected
        """
        self.match(IfDataTokenType.BEGIN)
        tk = self.current_token
        tk_type = tk.type
        tk_value = tk.value

        if tk_type == IfDataTokenType.IDENT:
            self.consume()
            result = self.enter(self.syntax_tos.type)
            self.leave()
            self.match(IfDataTokenType.END)
            self.match(IfDataTokenType.IDENT, tk_value)
            return result
        else:
            raise TypeError(f"Expected IDENT got {tk_type}[{tk_value!r}].")

    def tagged_struct(self) -> Dict:
        """
        Parse a tagged struct in the IF_DATA section.

        Returns:
            A dictionary representing the parsed tagged struct
        """
        result = defaultdict(list)

        while True:
            # Check for end of struct
            if self.current_token.type == IfDataTokenType.END:
                break

            # Get the token to check
            if self.current_token.type == IfDataTokenType.BEGIN:
                tk = self.lookahead(1)
            else:
                tk = self.current_token

            tk_type = tk.type
            tk_value = tk.value

            # Check if the token is a valid member
            if tk_value not in self.syntax_tos.members:
                break

            if tk_type == IfDataTokenType.IDENT:
                elem = self.syntax_tos.members.get(tk_value)

                # Handle flag values with no further definition
                if (
                    isinstance(elem, TaggedStructMember)
                    and not isinstance(elem.definition, Block)
                    and elem.definition.member is None
                ):
                    result[tk_value].append(True)
                    self.consume()
                else:
                    multiple = elem.multiple

                    # Handle multiple values
                    while True:
                        tmp_value = self.enter(elem.definition)
                        if tmp_value:
                            result[tk_value].append(tmp_value)
                        self.leave()

                        if not multiple:
                            break

                        if self.current_token.value != tk_value:
                            break

                        if self.current_token.type == IfDataTokenType.END:
                            break
            else:
                token_type_name = IfDataTokenType(tk.type).name
                self.logger.error(f"Invalid token {token_type_name} for tagged struct. Expected IDENT.")

        # Process the results
        return_value = {}
        for k, v in result.items():
            if isinstance(self.syntax_tos.members, dict):
                member = self.syntax_tos.members.get(k)
                if member is None:
                    return_value[k] = v
                else:
                    if not member.multiple:
                        return_value[k] = v[0]
                    else:
                        return_value[k] = v
            else:
                return_value[k] = v

        return return_value

    def tagged_union(self) -> Dict:
        """
        Parse a tagged union in the IF_DATA section.

        Returns:
            A dictionary representing the parsed tagged union
        """
        # Check if this is a block
        if self.current_token.type == IfDataTokenType.BEGIN:
            tk = self.lookahead(1)
            self.consume()
            block = True
        else:
            tk = self.current_token
            block = False

        # Validate token type
        if tk.type != IfDataTokenType.IDENT:
            token_type_name = IfDataTokenType(tk.type).name
            self.logger.error(f"Invalid token {token_type_name} for tagged union. Expected IDENT.")
            if block:
                self.rewind()
            return {}

        tk_value = tk.value
        self.consume()

        # Get the member
        mem_dict = self.syntax_tos.members
        if tk_value in mem_dict:
            member = mem_dict[tk_value]

            # Handle block members
            if isinstance(member, Block):
                result = self.enter(member.type)
                self.leave()
                self.match(IfDataTokenType.END)
                self.match(IfDataTokenType.IDENT, tk_value)
                return {tk_value: result}
            else:
                # Handle other members
                if isinstance(member, NullObject):
                    result = True  # Just a flag value with no further definition
                else:
                    result = self.enter(member.node)
                    self.leave()
                return {tk_value: result}
        else:
            self.logger.error(f"TAGGED_UNION has no member {tk_value!r}.")
            amount = 2 if block else 1
            self.rewind(amount)
            return {}

    def struct(self) -> List:
        """
        Parse a struct in the IF_DATA section.

        Returns:
            A list representing the parsed struct
        """
        result = []

        for member in self.syntax_tos.members:
            tmp_res = self.enter(member.node)
            if tmp_res != {}:
                result.append(tmp_res)
            self.leave()

        return result

    def validate_pdt(self, token: IfDataToken) -> bool:
        """
        Validate a predefined type token.

        Args:
            token: The token to validate

        Returns:
            True if the token is valid, False otherwise
        """
        pdt_type = self.syntax_tos.type

        # Validate token type against PDT type
        if pdt_type == AMLPredefinedTypeEnum.INT or pdt_type == AMLPredefinedTypeEnum.LONG:
            return token.type == IfDataTokenType.INT
        elif pdt_type == AMLPredefinedTypeEnum.FLOAT or pdt_type == AMLPredefinedTypeEnum.DOUBLE:
            return token.type in (IfDataTokenType.FLOAT, IfDataTokenType.INT)
        elif pdt_type == AMLPredefinedTypeEnum.CHAR:
            return token.type == IfDataTokenType.STRING and len(token.value) == 1

        # Default validation
        return True

    def pdt(self) -> Any:
        """
        Parse a predefined type in the IF_DATA section.

        Returns:
            The value of the predefined type
        """
        tk = self.current_token

        # Validate the token
        #
        # TODO: Add support for array types.
        # if not self.validate_pdt(tk):
        #    token_type_name = IfDataTokenType(tk.type).name
        #    pdt_type_name = AMLPredefinedTypeEnum(self.syntax_tos.type).name
        #    self.logger.warn(f"Token {token_type_name}[{tk.value}] may not be valid for PDT type {pdt_type_name}")

        self.consume()

        # Handle array specifications
        arr_spec = self.syntax_tos.arr_spec
        if arr_spec:
            # Array handling would go here
            pass

        return tk.value

    def enumeration(self) -> str:
        """
        Parse an enumeration in the IF_DATA section.

        Returns:
            The value of the enumeration
        """
        tk = self.current_token
        self.consume()

        # Validate the enumeration value
        if tk.value not in self.syntax_tos.values:
            self.logger.warn(f"Enumeration value {tk.value!r} not found in {self.syntax_tos.name}")

        return tk.value

    def member(self) -> Any:
        """
        Parse a member in the IF_DATA section.

        Returns:
            The parsed member
        """
        result = self.enter(self.syntax_tos.node)
        self.leave()
        return result

    def tagged_struct_definition(self) -> Union[List, Any]:
        """
        Parse a tagged struct definition in the IF_DATA section.

        Returns:
            The parsed tagged struct definition
        """
        result = None
        multiple = self.syntax_tos.multiple
        self.consume()

        if multiple:
            result = []
            while True:
                value = self.enter(self.syntax_tos.member)
                self.leave()
                result.append(value)
                if self.current_token.type in (IfDataTokenType.IDENT, IfDataTokenType.BEGIN, IfDataTokenType.END):
                    break
        else:
            result = self.enter(self.syntax_tos.member)
            self.leave()

        return result

    def enter(self, klass) -> Any:
        """
        Enter a node in the syntax tree.

        Args:
            klass: The node to enter

        Returns:
            The parsed result

        Raises:
            TypeError: If the node type is not supported
        """
        # Resolve references
        if isinstance(klass, Referrer):
            klass = self.ref_dict[klass.category][klass.identifier]

        self.syntax_stack.append(klass)

        try:
            # Dispatch based on node type
            if isinstance(klass, Block):
                result = self.block()
            elif isinstance(klass, TaggedStruct):
                result = self.tagged_struct()
            elif isinstance(klass, TaggedUnion):
                result = self.tagged_union()
            elif isinstance(klass, Struct):
                result = self.struct()
            elif isinstance(klass, Enumeration):
                result = self.enumeration()
            elif isinstance(klass, Member):
                result = self.member()
            elif isinstance(klass, TaggedStructDefinition):
                result = self.tagged_struct_definition()
            elif isinstance(klass, PDT):
                result = self.pdt()
            elif isinstance(klass, NullObject):
                result = {}
            else:
                self.logger.error(f"Unsupported class {type(klass).__name__}")
                raise TypeError(f"Unsupported class {type(klass).__name__}")

            self.level += 1
            return result
        except Exception as e:
            self.logger.error(f"Error in enter({type(klass).__name__}): {str(e)}")
            self.level += 1
            return {}

    def leave(self) -> None:
        """Leave the current node in the syntax tree."""
        self.syntax_stack.pop()
        self.level -= 1

    @property
    def syntax_tos(self) -> Any:
        """
        Get the top of the syntax stack.

        Returns:
            The top element of the syntax stack
        """
        return self.syntax_stack[-1]

    @property
    def current_token(self) -> IfDataToken:
        """
        Get the token at the current stream position.

        Returns:
            The current token
        """
        return self.lookahead(0)

    def lookahead(self, n: int = 1) -> IfDataToken:
        """
        Get the token `n` elements ahead of current stream position.

        Args:
            n: The number of elements to look ahead

        Returns:
            The token at position current + n

        Raises:
            EOFReached: If the position is beyond the end of the token stream
        """
        index = self.token_idx + n
        if index < self.num_tokens:
            return self.tokens[index]
        else:
            raise EOFReached()

    def consume(self) -> None:
        """Increment token stream position by one."""
        self.token_idx += 1

    def rewind(self, n: int = 1) -> None:
        """
        Back up token stream position by `n` elements.

        Args:
            n: The number of elements to back up
        """
        self.token_idx = max(0, self.token_idx - n)

    def match(self, token_type: IfDataTokenType, value: Optional[Any] = None) -> bool:
        """
        Match the current token against the expected type and value.

        Args:
            token_type: The expected token type
            value: The expected token value (optional)

        Returns:
            True if the match succeeded, False otherwise
        """
        try:
            ok = self.current_token.type == token_type

            if not ok:
                expected_type_name = token_type.name
                actual_type_name = IfDataTokenType(self.current_token.type).name
                self.logger.error(f"{expected_type_name} not matched against {actual_type_name}")

                # Try to recover by finding the next token of the expected type
                recovery_limit = 10  # Limit recovery attempts to avoid infinite loops
                recovery_count = 0

                while recovery_count < recovery_limit:
                    self.consume()
                    recovery_count += 1

                    if self.current_token.type == token_type:
                        ok = True
                        break

                if not ok:
                    self.logger.error(f"Failed to recover from token mismatch after {recovery_count} attempts")

            token_value = self.current_token.value
            self.consume()

            if value is None:
                return ok
            else:
                if not ok:
                    return False

                value_match = token_value == value
                if not value_match:
                    self.logger.error(f"Value mismatch: expected {value!r}, got {token_value!r}")

                return value_match

        except EOFReached:
            self.logger.error("Unexpected end of input during token matching")
            return False

    def traverse(self, node) -> Any:
        """
        Traverse the AML tree and convert it to the internal representation.

        Args:
            node: The AML node to traverse

        Returns:
            The internal representation of the node
        """
        result = None

        try:
            if node.aml_type == AmlType.ROOT:
                result = Root()
                members = node.map.get("MEMBERS").list
                for mem in members:
                    result.members.append(self.traverse(mem))

            elif node.aml_type == AmlType.STRUCT:
                result = Struct()
                mp = node.map
                name = mp.get("NAME")
                result.name = name.value
                members = mp.get("MEMBERS").list
                for mem in members:
                    result.members.append(self.traverse(mem))

            elif node.aml_type == AmlType.MEMBER:
                mp = node.map
                is_block = bool(mp.get("IS_BLOCK").value)
                tmp_node = self.traverse(mp.get("NODE"))
                result = Member(tmp_node, is_block)

            elif node.aml_type == AmlType.STRUCT_MEMBER:
                member = node.map.get("MEMBER")
                is_block = bool(member.map.get("IS_BLOCK").value)
                tmp_node = self.traverse(member.map.get("NODE"))
                result = Member(tmp_node, is_block)

            elif node.aml_type == AmlType.MEMBERS:
                result = []
                for mem in node.list:
                    result.append(self.traverse(mem))

            elif node.aml_type == AmlType.PDT:
                mp = node.map
                result = PDT(AMLPredefinedTypeEnum(mp.get("TYPE").value), [a.value for a in mp.get("ARR_SPEC").list])

            elif node.aml_type == AmlType.ENUMERATION:
                result = Enumeration()
                mp = node.map
                name = mp.get("NAME")
                result.name = name.value
                values = mp.get("VALUES").list
                for value in values:
                    result.values[value.map.get("NAME").value] = value.map.get("VALUE").value

            elif node.aml_type == AmlType.TAGGED_STRUCT:
                result = TaggedStruct()
                mp = node.map
                name = mp.get("NAME")
                result.name = name.value
                members = mp.get("MEMBERS").list
                for mem in members:
                    tag = mem.map["TAG"].value
                    result.members[tag] = self.traverse(mem.map["MEMBER"])

            elif node.aml_type == AmlType.TAGGED_STRUCT_MEMBER:
                mp = node.map
                result = TaggedStructMember(self.traverse(mp.get("DEFINITION")), bool(mp.get("MULTIPLE").value))

            elif node.aml_type == AmlType.TAGGED_STRUCT_DEFINITION:
                mp = node.map
                result = TaggedStructDefinition(self.traverse(mp.get("MEMBER")), bool(mp.get("MULTIPLE").value))

            elif node.aml_type == AmlType.TAGGED_UNION:
                result = TaggedUnion()
                mp = node.map
                name = mp.get("NAME")
                result.name = name.value
                members = mp.get("MEMBERS").list
                for mem in members:
                    tag = mem.map["TAG"].value
                    result.members[tag] = self.traverse(mem.map["MEMBER"])

            elif node.aml_type == AmlType.TAGGED_UNION_MEMBER:
                # This case seems incomplete in the original code
                mp = node.map
                tag = mp.get("TAG").value
                member = mp.get("MEMBER")
                # Return a proper result for this case
                result = self.traverse(member)

            elif node.aml_type == AmlType.BLOCK:
                mp = node.map
                tag = mp.get("TAG").value
                tp = self.traverse(mp.get("TYPE"))
                result = Block(tag, tp)

            elif node.aml_type == AmlType.REFERRER:
                result = Referrer(ReferrerType(node.map.get("CATEGORY").value), node.map.get("IDENTIFIER").value)

            elif node.aml_type == AmlType.NULL_NODE:
                result = NullObject()

            elif node.aml_type == AmlType.NONE:
                result = None

            else:
                self.logger.warn(f"Unhandled AML type: {node.aml_type}")

        except Exception as e:
            self.logger.error(f"Error traversing node of type {getattr(node, 'aml_type', 'unknown')}: {str(e)}")

        return result



import sys
from pprint import pprint

from pya2l import DB
from pya2l.api.inspect import Project


db = DB()
session = db.open_create(sys.argv[1], loglevel="DEBUG")
prj = Project(session)

mod = prj.module[0]
pprint(mod.variant_coding)



#!/usr/bin/env python
# -*- coding: latin-1 -*-
import sys

import pytest

from pya2l import DB
from pya2l.cse_units import CSE
from pya2l.cse_units import CSE_Type
from pya2l.cse_units import Referer

if sys.platform == "win32":
    FILE_NAME = r"f:\projects\examples\example-a2l-file"
else:
    FILE_NAME = r"~/projects/examples/example-a2l-file"  # We're assuming a Unix-like OS.


@pytest.mark.skip
def test_filename_with_a2l_ext():
    db = DB()
    db._set_path_components(FILE_NAME)

    assert db._pth == r"f:\projects\examples"
    assert db._base == r"example-a2l-file.a2l"
    assert db._dbfn == r"example-a2l-file.a2ldb"
    assert db._a2lfn == r"example-a2l-file.a2l"


@pytest.mark.skip
def test_filename_without_ext():
    db = DB()
    db._set_path_components(FILE_NAME)

    assert db._pth == r"f:\projects\examples"
    assert db._base == r"example-a2l-file"
    assert db._dbfn == r"example-a2l-file.a2ldb"
    assert db._a2lfn == r"example-a2l-file.a2l"


@pytest.mark.skip
def test_filename_with_a2ldb_ext():
    db = DB()
    db._set_path_components(FILE_NAME)

    assert db._pth == r"f:\projects\examples"
    assert db._base == r"example-a2l-file.a2ldb"
    assert db._dbfn == r"example-a2l-file.a2ldb"
    assert db._a2lfn == r"example-a2l-file.a2l"


@pytest.mark.skip
def test_filename_with_arbitrary_ext():
    db = DB()
    db._set_path_components(FILE_NAME)

    assert db._pth == r"f:\projects\examples"
    assert db._base == r"example-a2l-file.foobar"
    assert db._dbfn == r"example-a2l-file.a2ldb"
    assert db._a2lfn == r"example-a2l-file.foobar"


def test_cse_units():
    assert CSE.get(2) == CSE_Type(2, "100 µsec", Referer.TIME, "")



#!/usr/bin/env python
"""Classes for creating new instances of A2L database entities.

This module provides classes and functions to create new instances of entities
defined in the A2L database, such as Measurement, AxisPts, CompuMethod, etc.
It complements the inspect.py module, which provides read-only access to the database.

The module includes creator classes for all major A2L entities, with methods to
create instances and set their properties. Each creator class follows a consistent
pattern with a main creation method and additional methods to set optional attributes.
"""
__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2020-2025 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from typing import Any, Dict, List, Optional, Tuple, Union

import pya2l.model as model
from pya2l import exceptions
from pya2l.api.inspect import (  # Project, Header, Annotation
    AxisDescr,
    AxisPts,
    Characteristic,
    CompuMethod,
    Function,
    Group,
    Instance,
    Measurement,
    ModCommon,
    ModPar,
    NoCompuMethod,
    RecordLayout,
    StructureComponent,
    TypedefCharacteristic,
    TypedefMeasurement,
    TypedefStructure,
    VariantCoding,
)


class Creator:
    """Base class for all creator classes in this module.

    This class provides common functionality for creating new instances of
    database entities.

    Attributes
    ----------
    session : Any
        SQLAlchemy session object used to interact with the database
    """

    def __init__(self, session: Any):
        """Initialize a Creator instance.

        Parameters
        ----------
        session : Any
            SQLAlchemy session object used to interact with the database
        """
        self.session = session

    def commit(self):
        """Commit the current transaction to the database."""
        self.session.commit()

    def rollback(self):
        """Rollback the current transaction."""
        self.session.rollback()


class CompuMethodCreator(Creator):
    """Creator class for CompuMethod entities.

    This class provides methods to create new CompuMethod instances and
    their related entities (Coeffs, CoeffsLinear, Formula, etc.).

    Attributes
    ----------
    session : Any
        SQLAlchemy session object used to interact with the database
    """

    def create_compu_method(
        self, name: str, long_identifier: str, conversion_type: str, format_str: str, unit: str, module_name: Optional[str] = None
    ) -> model.CompuMethod:
        """Create a new CompuMethod instance.

        Parameters
        ----------
        name : str
            Name of the computation method
        long_identifier : str
            Description of the computation method
        conversion_type : str
            Type of conversion (e.g., "IDENTICAL", "LINEAR", "RAT_FUNC", "TAB_INTP")
        format_str : str
            Format string for displaying physical values
        unit : str
            Physical unit of the values
        module_name : Optional[str], optional
            Name of the module to associate with this CompuMethod, by default None

        Returns
        -------
        model.CompuMethod
            The newly created CompuMethod instance
        """
        # Find the module if module_name is provided
        module = None
        if module_name:
            module = self.session.query(model.Module).filter(model.Module.name == module_name).first()
            if not module:
                raise exceptions.StructuralError(f"Module '{module_name}' not found")

        # Create the CompuMethod instance
        compu_method = model.CompuMethod(
            name=name, longIdentifier=long_identifier, conversionType=conversion_type, format=format_str, unit=unit
        )

        # Associate with module if provided
        if module:
            compu_method.module = module

        # Add to session
        self.session.add(compu_method)

        return compu_method

    def add_coeffs(
        self, compu_method: model.CompuMethod, a: float, b: float, c: float, d: float, e: float, f: float
    ) -> model.Coeffs:
        """Add Coeffs to a CompuMethod.

        Parameters
        ----------
        compu_method : model.CompuMethod
            The CompuMethod to add Coeffs to
        a : float
            Coefficient a
        b : float
            Coefficient b
        c : float
            Coefficient c
        d : float
            Coefficient d
        e : float
            Coefficient e
        f : float
            Coefficient f

        Returns
        -------
        model.Coeffs
            The newly created Coeffs instance
        """
        coeffs = model.Coeffs(a=a, b=b, c=c, d=d, e=e, f=f)
        coeffs.compu_method = compu_method
        self.session.add(coeffs)
        return coeffs

    def add_coeffs_linear(self, compu_method: model.CompuMethod, a: float, b: float) -> model.CoeffsLinear:
        """Add CoeffsLinear to a CompuMethod.

        Parameters
        ----------
        compu_method : model.CompuMethod
            The CompuMethod to add CoeffsLinear to
        a : float
            Coefficient a
        b : float
            Coefficient b

        Returns
        -------
        model.CoeffsLinear
            The newly created CoeffsLinear instance
        """
        coeffs_linear = model.CoeffsLinear(a=a, b=b)
        coeffs_linear.compu_method = compu_method
        self.session.add(coeffs_linear)
        return coeffs_linear

    def add_compu_tab_ref(self, compu_method: model.CompuMethod, conversion_table: str) -> model.CompuTabRef:
        """Add CompuTabRef to a CompuMethod.

        Parameters
        ----------
        compu_method : model.CompuMethod
            The CompuMethod to add CompuTabRef to
        conversion_table : str
            Name of the conversion table

        Returns
        -------
        model.CompuTabRef
            The newly created CompuTabRef instance
        """
        compu_tab_ref = model.CompuTabRef(conversionTable=conversion_table)
        compu_tab_ref.compu_method = compu_method
        self.session.add(compu_tab_ref)
        return compu_tab_ref

    def add_formula(self, compu_method: model.CompuMethod, formula_str: str, formula_inv: Optional[str] = None) -> model.Formula:
        """Add Formula to a CompuMethod.

        Parameters
        ----------
        compu_method : model.CompuMethod
            The CompuMethod to add Formula to
        formula_str : str
            The formula string (f_x)
        formula_inv : Optional[str], optional
            The inverse formula string, by default None

        Returns
        -------
        model.Formula
            The newly created Formula instance
        """
        formula = model.Formula(f_x=formula_str)
        formula.compu_method = compu_method
        self.session.add(formula)

        # Add inverse formula if provided
        if formula_inv:
            formula_inv_obj = model.FormulaInv(g_x=formula_inv)
            formula_inv_obj.formula = formula
            self.session.add(formula_inv_obj)

        return formula

    def add_ref_unit(self, compu_method: model.CompuMethod, unit: str) -> model.RefUnit:
        """Add RefUnit to a CompuMethod.

        Parameters
        ----------
        compu_method : model.CompuMethod
            The CompuMethod to add RefUnit to
        unit : str
            The reference unit

        Returns
        -------
        model.RefUnit
            The newly created RefUnit instance
        """
        ref_unit = model.RefUnit(unit=unit)
        ref_unit.compu_method = compu_method
        self.session.add(ref_unit)
        return ref_unit

    def add_status_string_ref(self, compu_method: model.CompuMethod, conversion_table: str) -> model.StatusStringRef:
        """Add StatusStringRef to a CompuMethod.

        Parameters
        ----------
        compu_method : model.CompuMethod
            The CompuMethod to add StatusStringRef to
        conversion_table : str
            Name of the conversion table

        Returns
        -------
        model.StatusStringRef
            The newly created StatusStringRef instance
        """
        status_string_ref = model.StatusStringRef(conversionTable=conversion_table)
        status_string_ref.compu_method = compu_method
        self.session.add(status_string_ref)
        return status_string_ref


class MeasurementCreator(Creator):
    """Creator class for Measurement entities.

    This class provides methods to create new Measurement instances and
    their related entities (ArraySize, BitOperation, etc.).

    Attributes
    ----------
    session : Any
        SQLAlchemy session object used to interact with the database
    """

    def create_measurement(
        self,
        name: str,
        long_identifier: str,
        datatype: str,
        conversion: str,
        resolution: int,
        accuracy: float,
        lower_limit: float,
        upper_limit: float,
        module_name: Optional[str] = None,
    ) -> model.Measurement:
        """Create a new Measurement instance.

        Parameters
        ----------
        name : str
            Name of the measurement
        long_identifier : str
            Description of the measurement
        datatype : str
            Data type of the measurement (e.g., "UBYTE", "SWORD", "FLOAT32_IEEE")
        conversion : str
            Name of the computation method for conversion
        resolution : int
            Resolution of the measurement
        accuracy : float
            Accuracy of the measurement
        lower_limit : float
            Lower limit of the measurement
        upper_limit : float
            Upper limit of the measurement
        module_name : Optional[str], optional
            Name of the module to associate with this Measurement, by default None

        Returns
        -------
        model.Measurement
            The newly created Measurement instance
        """
        # Find the module if module_name is provided
        module = None
        if module_name:
            module = self.session.query(model.Module).filter(model.Module.name == module_name).first()
            if not module:
                raise exceptions.StructuralError(f"Module '{module_name}' not found")

        # Create the Measurement instance
        measurement = model.Measurement(
            name=name,
            longIdentifier=long_identifier,
            datatype=datatype,
            conversion=conversion,
            resolution=resolution,
            accuracy=accuracy,
            lowerLimit=lower_limit,
            upperLimit=upper_limit,
        )

        # Associate with module if provided
        if module:
            measurement.module = module

        # Add to session
        self.session.add(measurement)

        return measurement

    def add_array_size(self, measurement: model.Measurement, number: int) -> model.ArraySize:
        """Add ArraySize to a Measurement.

        Parameters
        ----------
        measurement : model.Measurement
            The Measurement to add ArraySize to
        number : int
            The array size

        Returns
        -------
        model.ArraySize
            The newly created ArraySize instance
        """
        array_size = model.ArraySize(number=number)
        array_size.measurement = measurement
        self.session.add(array_size)
        return array_size

    def add_bit_operation(
        self,
        measurement: model.Measurement,
        left_shift: Optional[int] = None,
        right_shift: Optional[int] = None,
        sign_extend: Optional[bool] = None,
    ) -> model.BitOperation:
        """Add BitOperation to a Measurement.

        Parameters
        ----------
        measurement : model.Measurement
            The Measurement to add BitOperation to
        left_shift : Optional[int], optional
            Left shift value, by default None
        right_shift : Optional[int], optional
            Right shift value, by default None
        sign_extend : Optional[bool], optional
            Sign extend flag, by default None

        Returns
        -------
        model.BitOperation
            The newly created BitOperation instance
        """
        bit_operation = model.BitOperation()
        bit_operation.measurement = measurement
        self.session.add(bit_operation)

        # Add left shift if provided
        if left_shift is not None:
            left_shift_obj = model.LeftShift(bitCount=left_shift)
            left_shift_obj.bit_operation = bit_operation
            self.session.add(left_shift_obj)

        # Add right shift if provided
        if right_shift is not None:
            right_shift_obj = model.RightShift(bitCount=right_shift)
            right_shift_obj.bit_operation = bit_operation
            self.session.add(right_shift_obj)

        # Add sign extend if provided
        if sign_extend is not None:
            sign_extend_obj = model.SignExtend()
            sign_extend_obj.bit_operation = bit_operation
            self.session.add(sign_extend_obj)

        return bit_operation

    def add_ecu_address(self, measurement: model.Measurement, address: int) -> model.EcuAddress:
        """Add EcuAddress to a Measurement.

        Parameters
        ----------
        measurement : model.Measurement
            The Measurement to add EcuAddress to
        address : int
            The ECU address

        Returns
        -------
        model.EcuAddress
            The newly created EcuAddress instance
        """
        ecu_address = model.EcuAddress(address=address)
        ecu_address.measurement = measurement
        self.session.add(ecu_address)
        return ecu_address

    def add_matrix_dim(
        self, measurement: model.Measurement, x_dim: int, y_dim: Optional[int] = None, z_dim: Optional[int] = None
    ) -> model.MatrixDim:
        """Add MatrixDim to a Measurement.

        Parameters
        ----------
        measurement : model.Measurement
            The Measurement to add MatrixDim to
        x_dim : int
            X dimension
        y_dim : Optional[int], optional
            Y dimension, by default None
        z_dim : Optional[int], optional
            Z dimension, by default None

        Returns
        -------
        model.MatrixDim
            The newly created MatrixDim instance
        """
        matrix_dim = model.MatrixDim(xDim=x_dim, yDim=y_dim if y_dim is not None else 0, zDim=z_dim if z_dim is not None else 0)
        matrix_dim.measurement = measurement
        self.session.add(matrix_dim)
        return matrix_dim

    def add_format(self, measurement: model.Measurement, format_string: str) -> model.Format:
        """Add Format to a Measurement.

        Parameters
        ----------
        measurement : model.Measurement
            The Measurement to add Format to
        format_string : str
            The format string

        Returns
        -------
        model.Format
            The newly created Format instance
        """
        format_obj = model.Format(formatString=format_string)
        format_obj.measurement = measurement
        self.session.add(format_obj)
        return format_obj

    def add_phys_unit(self, measurement: model.Measurement, unit: str) -> model.PhysUnit:
        """Add PhysUnit to a Measurement.

        Parameters
        ----------
        measurement : model.Measurement
            The Measurement to add PhysUnit to
        unit : str
            The physical unit

        Returns
        -------
        model.PhysUnit
            The newly created PhysUnit instance
        """
        phys_unit = model.PhysUnit(unit=unit)
        phys_unit.measurement = measurement
        self.session.add(phys_unit)
        return phys_unit

    def add_symbol_link(self, measurement: model.Measurement, symbol_name: str, offset: Optional[int] = None) -> model.SymbolLink:
        """Add SymbolLink to a Measurement.

        Parameters
        ----------
        measurement : model.Measurement
            The Measurement to add SymbolLink to
        symbol_name : str
            The symbol name
        offset : Optional[int], optional
            The offset, by default None

        Returns
        -------
        model.SymbolLink
            The newly created SymbolLink instance
        """
        symbol_link = model.SymbolLink(symbolName=symbol_name, offset=offset if offset is not None else 0)
        symbol_link.measurement = measurement
        self.session.add(symbol_link)
        return symbol_link

    def add_virtual(self, measurement: model.Measurement, measuring_channel: Optional[List[str]] = None) -> model.Virtual:
        """Add Virtual to a Measurement.

        Parameters
        ----------
        measurement : model.Measurement
            The Measurement to add Virtual to
        measuring_channel : Optional[List[str]], optional
            List of measuring channels, by default None

        Returns
        -------
        model.Virtual
            The newly created Virtual instance
        """
        virtual = model.Virtual()
        virtual.measurement = measurement
        self.session.add(virtual)

        # Add measuring channels if provided
        if measuring_channel:
            for channel in measuring_channel:
                channel_obj = model.VirtualMeasuringChannels(measuringChannel=channel)
                channel_obj.virtual = virtual
                self.session.add(channel_obj)

        return virtual


class AxisPtsCreator(Creator):
    """Creator class for AxisPts entities.

    This class provides methods to create new AxisPts instances and
    their related entities.

    Attributes
    ----------
    session : Any
        SQLAlchemy session object used to interact with the database
    """

    def create_axis_pts(
        self,
        name: str,
        long_identifier: str,
        address: int,
        input_quantity: str,
        deposit_attr: str,
        max_diff: float,
        conversion: str,
        max_axis_points: int,
        lower_limit: float,
        upper_limit: float,
        module_name: Optional[str] = None,
    ) -> model.AxisPts:
        """Create a new AxisPts instance.

        Parameters
        ----------
        name : str
            Name of the axis points
        long_identifier : str
            Description of the axis points
        address : int
            Address of the axis points
        input_quantity : str
            Input quantity of the axis points
        deposit_attr : str
            Deposit attribute of the axis points
        max_diff : float
            Maximum difference of the axis points
        conversion : str
            Name of the computation method for conversion
        max_axis_points : int
            Maximum number of axis points
        lower_limit : float
            Lower limit of the axis points
        upper_limit : float
            Upper limit of the axis points
        module_name : Optional[str], optional
            Name of the module to associate with this AxisPts, by default None

        Returns
        -------
        model.AxisPts
            The newly created AxisPts instance
        """
        # Find the module if module_name is provided
        module = None
        if module_name:
            module = self.session.query(model.Module).filter(model.Module.name == module_name).first()
            if not module:
                raise exceptions.StructuralError(f"Module '{module_name}' not found")

        # Create the AxisPts instance
        axis_pts = model.AxisPts(
            name=name,
            longIdentifier=long_identifier,
            address=address,
            inputQuantity=input_quantity,
            depositAttr=deposit_attr,
            maxDiff=max_diff,
            conversion=conversion,
            maxAxisPoints=max_axis_points,
            lowerLimit=lower_limit,
            upperLimit=upper_limit,
        )

        # Associate with module if provided
        if module:
            axis_pts.module = module

        # Add to session
        self.session.add(axis_pts)

        return axis_pts

    def add_byte_order(self, axis_pts: model.AxisPts, byte_order: str) -> model.ByteOrder:
        """Add ByteOrder to an AxisPts.

        Parameters
        ----------
        axis_pts : model.AxisPts
            The AxisPts to add ByteOrder to
        byte_order : str
            The byte order (e.g., "MSB_FIRST", "MSB_LAST")

        Returns
        -------
        model.ByteOrder
            The newly created ByteOrder instance
        """
        byte_order_obj = model.ByteOrder(byteOrder=byte_order)
        byte_order_obj.axis_pts = axis_pts
        self.session.add(byte_order_obj)
        return byte_order_obj

    def add_calibration_access(self, axis_pts: model.AxisPts, calibration_access: str) -> model.CalibrationAccess:
        """Add CalibrationAccess to an AxisPts.

        Parameters
        ----------
        axis_pts : model.AxisPts
            The AxisPts to add CalibrationAccess to
        calibration_access : str
            The calibration access (e.g., "CALIBRATION", "NO_CALIBRATION")

        Returns
        -------
        model.CalibrationAccess
            The newly created CalibrationAccess instance
        """
        calibration_access_obj = model.CalibrationAccess(type=calibration_access)
        calibration_access_obj.axis_pts = axis_pts
        self.session.add(calibration_access_obj)
        return calibration_access_obj

    def add_deposit(self, axis_pts: model.AxisPts, mode: str) -> model.Deposit:
        """Add Deposit to an AxisPts.

        Parameters
        ----------
        axis_pts : model.AxisPts
            The AxisPts to add Deposit to
        mode : str
            The deposit mode (e.g., "ABSOLUTE", "DIFFERENCE")

        Returns
        -------
        model.Deposit
            The newly created Deposit instance
        """
        deposit = model.Deposit(mode=mode)
        deposit.axis_pts = axis_pts
        self.session.add(deposit)
        return deposit

    def add_extended_limits(self, axis_pts: model.AxisPts, lower_limit: float, upper_limit: float) -> model.ExtendedLimits:
        """Add ExtendedLimits to an AxisPts.

        Parameters
        ----------
        axis_pts : model.AxisPts
            The AxisPts to add ExtendedLimits to
        lower_limit : float
            The lower limit
        upper_limit : float
            The upper limit

        Returns
        -------
        model.ExtendedLimits
            The newly created ExtendedLimits instance
        """
        extended_limits = model.ExtendedLimits(lowerLimit=lower_limit, upperLimit=upper_limit)
        extended_limits.axis_pts = axis_pts
        self.session.add(extended_limits)
        return extended_limits

    def add_format(self, axis_pts: model.AxisPts, format_string: str) -> model.Format:
        """Add Format to an AxisPts.

        Parameters
        ----------
        axis_pts : model.AxisPts
            The AxisPts to add Format to
        format_string : str
            The format string

        Returns
        -------
        model.Format
            The newly created Format instance
        """
        format_obj = model.Format(formatString=format_string)
        format_obj.axis_pts = axis_pts
        self.session.add(format_obj)
        return format_obj

    def add_monotony(self, axis_pts: model.AxisPts, monotony: str) -> model.Monotony:
        """Add Monotony to an AxisPts.

        Parameters
        ----------
        axis_pts : model.AxisPts
            The AxisPts to add Monotony to
        monotony : str
            The monotony (e.g., "MON_INCREASE", "MON_DECREASE", "STRICT_INCREASE", "STRICT_DECREASE")

        Returns
        -------
        model.Monotony
            The newly created Monotony instance
        """
        monotony_obj = model.Monotony(monotony=monotony)
        monotony_obj.axis_pts = axis_pts
        self.session.add(monotony_obj)
        return monotony_obj

    def add_phys_unit(self, axis_pts: model.AxisPts, unit: str) -> model.PhysUnit:
        """Add PhysUnit to an AxisPts.

        Parameters
        ----------
        axis_pts : model.AxisPts
            The AxisPts to add PhysUnit to
        unit : str
            The physical unit

        Returns
        -------
        model.PhysUnit
            The newly created PhysUnit instance
        """
        phys_unit = model.PhysUnit(unit=unit)
        phys_unit.axis_pts = axis_pts
        self.session.add(phys_unit)
        return phys_unit

    def add_step_size(self, axis_pts: model.AxisPts, step_size: float) -> model.StepSize:
        """Add StepSize to an AxisPts.

        Parameters
        ----------
        axis_pts : model.AxisPts
            The AxisPts to add StepSize to
        step_size : float
            The step size

        Returns
        -------
        model.StepSize
            The newly created StepSize instance
        """
        step_size_obj = model.StepSize(stepSize=step_size)
        step_size_obj.axis_pts = axis_pts
        self.session.add(step_size_obj)
        return step_size_obj

    def add_symbol_link(self, axis_pts: model.AxisPts, symbol_name: str, offset: Optional[int] = None) -> model.SymbolLink:
        """Add SymbolLink to an AxisPts.

        Parameters
        ----------
        axis_pts : model.AxisPts
            The AxisPts to add SymbolLink to
        symbol_name : str
            The symbol name
        offset : Optional[int], optional
            The offset, by default None

        Returns
        -------
        model.SymbolLink
            The newly created SymbolLink instance
        """
        symbol_link = model.SymbolLink(symbolName=symbol_name, offset=offset if offset is not None else 0)
        symbol_link.axis_pts = axis_pts
        self.session.add(symbol_link)
        return symbol_link


class ProjectCreator(Creator):
    """Creator class for Project entities.

    This class provides methods to create new Project instances and
    their related entities (Header, Module, etc.).

    Attributes
    ----------
    session : Any
        SQLAlchemy session object used to interact with the database
    """

    def create_project(self, name: str, long_identifier: str) -> model.Project:
        """Create a new Project instance.

        Parameters
        ----------
        name : str
            Name of the project
        long_identifier : str
            Description of the project

        Returns
        -------
        model.Project
            The newly created Project instance
        """
        # Create the Project instance
        project = model.Project(name=name, longIdentifier=long_identifier)

        # Add to session
        self.session.add(project)

        return project

    def add_header(self, project: model.Project, comment: str) -> model.Header:
        """Add Header to a Project.

        Parameters
        ----------
        project : model.Project
            The Project to add Header to
        comment : str
            The comment for the header

        Returns
        -------
        model.Header
            The newly created Header instance
        """
        header = model.Header(comment=comment)
        header.project = project
        self.session.add(header)
        return header

    def add_project_no(self, header: model.Header, project_number: str) -> model.ProjectNo:
        """Add ProjectNo to a Header.

        Parameters
        ----------
        header : model.Header
            The Header to add ProjectNo to
        project_number : str
            The project number

        Returns
        -------
        model.ProjectNo
            The newly created ProjectNo instance
        """
        project_no = model.ProjectNo(projectNumber=project_number)
        project_no.header = header
        self.session.add(project_no)
        return project_no


class ModuleCreator(Creator):
    """Creator class for Module entities.

    This class provides methods to create new Module instances and
    their related entities. It includes methods to create and associate
    various entities with a Module, such as measurements, characteristics,
    axis points, computation methods, etc.

    Attributes
    ----------
    session : Any
        SQLAlchemy session object used to interact with the database
    """

    def create_module(self, name: str, long_identifier: str, project: Optional[model.Project] = None) -> model.Module:
        """Create a new Module instance.

        Parameters
        ----------
        name : str
            Name of the module
        long_identifier : str
            Description of the module
        project : Optional[model.Project], optional
            The Project to associate with this Module, by default None

        Returns
        -------
        model.Module
            The newly created Module instance
        """
        # Create the Module instance
        module = model.Module(name=name, longIdentifier=long_identifier)

        # Associate with project if provided
        if project:
            module.project = project

        # Add to session
        self.session.add(module)

        return module

    def add_if_data(self, module: model.Module, if_data_text: str) -> model.IfData:
        """Add IF_DATA to a Module.

        Parameters
        ----------
        module : model.Module
            The Module to add IF_DATA to
        if_data_text : str
            The IF_DATA text content

        Returns
        -------
        model.IfData
            The newly created IfData instance
        """
        if_data = model.IfData(content=if_data_text)
        if_data.module = module
        self.session.add(if_data)
        return if_data

    def add_variant_coding(
        self, module: model.Module, var_separator: Optional[str] = None, var_naming: Optional[str] = None
    ) -> model.VariantCoding:
        """Add VariantCoding to a Module.

        Parameters
        ----------
        module : model.Module
            The Module to add VariantCoding to
        var_separator : Optional[str], optional
            The variant separator, by default None
        var_naming : Optional[str], optional
            The variant naming convention, by default None

        Returns
        -------
        model.VariantCoding
            The newly created VariantCoding instance
        """
        variant_coding = model.VariantCoding()
        variant_coding.module = module
        self.session.add(variant_coding)

        # Add var_separator if provided
        if var_separator:
            var_separator_obj = model.VarSeparator(separator=var_separator)
            var_separator_obj.variant_coding = variant_coding
            self.session.add(var_separator_obj)

        # Add var_naming if provided
        if var_naming:
            var_naming_obj = model.VarNaming(naming=var_naming)
            var_naming_obj.variant_coding = variant_coding
            self.session.add(var_naming_obj)

        return variant_coding

    def add_var_characteristic(
        self, variant_coding: model.VariantCoding, name: str, criterion_name: str, criterion_value: str
    ) -> model.VarCharacteristic:
        """Add VarCharacteristic to a VariantCoding.

        Parameters
        ----------
        variant_coding : model.VariantCoding
            The VariantCoding to add VarCharacteristic to
        name : str
            The name of the characteristic
        criterion_name : str
            The name of the criterion
        criterion_value : str
            The value of the criterion

        Returns
        -------
        model.VarCharacteristic
            The newly created VarCharacteristic instance
        """
        var_characteristic = model.VarCharacteristic(name=name)
        var_characteristic.variant_coding = variant_coding
        self.session.add(var_characteristic)

        # Add criterion
        criterion = model.VarCriterion(name=criterion_name, value=criterion_value)
        criterion.var_characteristic = var_characteristic
        self.session.add(criterion)

        return var_characteristic

    def add_user_rights(
        self, module: model.Module, user_level_id: str, read_only: bool = False, ref_group: Optional[str] = None
    ) -> model.UserRights:
        """Add UserRights to a Module.

        Parameters
        ----------
        module : model.Module
            The Module to add UserRights to
        user_level_id : str
            The user level ID
        read_only : bool, optional
            Whether the user rights are read-only, by default False
        ref_group : Optional[str], optional
            The reference group, by default None

        Returns
        -------
        model.UserRights
            The newly created UserRights instance
        """
        user_rights = model.UserRights(userLevelId=user_level_id, readOnly=read_only)
        user_rights.module = module
        self.session.add(user_rights)

        # Add ref_group if provided
        if ref_group:
            ref_group_obj = model.RefGroup(identifier=ref_group)
            ref_group_obj.user_rights = user_rights
            self.session.add(ref_group_obj)

        return user_rights

    def add_group(self, module: model.Module, group_name: str, long_identifier: str, root: bool = False) -> model.Group:
        """Add Group to a Module.

        Parameters
        ----------
        module : model.Module
            The Module to add Group to
        group_name : str
            The name of the group
        long_identifier : str
            The description of the group
        root : bool, optional
            Whether this is a root group, by default False

        Returns
        -------
        model.Group
            The newly created Group instance
        """
        group = model.Group(groupName=group_name, groupLongIdentifier=long_identifier)
        group.module = module
        self.session.add(group)

        # Add root if true
        if root:
            root_obj = model.Root()
            root_obj.group = group
            self.session.add(root_obj)

        return group

    def add_sub_group(self, group: model.Group, sub_group_name: str) -> model.SubGroup:
        """Add SubGroup to a Group.

        Parameters
        ----------
        group : model.Group
            The Group to add SubGroup to
        sub_group_name : str
            The name of the sub-group

        Returns
        -------
        model.SubGroup
            The newly created SubGroup instance
        """
        sub_group = model.SubGroup(identifier=sub_group_name)
        sub_group.group = group
        self.session.add(sub_group)
        return sub_group

    def add_unit(
        self,
        module: model.Module,
        name: str,
        long_identifier: str,
        display: str,
        type_str: Optional[str] = None,
        ref_unit: Optional[str] = None,
        unit_conversion: Optional[Dict[str, float]] = None,
    ) -> model.Unit:
        """Add Unit to a Module.

        Parameters
        ----------
        module : model.Module
            The Module to add Unit to
        name : str
            The name of the unit
        long_identifier : str
            The description of the unit
        display : str
            The display string for the unit
        type_str : Optional[str], optional
            The type of the unit, by default None
        ref_unit : Optional[str], optional
            The reference unit, by default None
        unit_conversion : Optional[Dict[str, float]], optional
            The unit conversion parameters (gradient, offset), by default None

        Returns
        -------
        model.Unit
            The newly created Unit instance
        """
        unit = model.Unit(name=name, longIdentifier=long_identifier, display=display, type=type_str)
        unit.module = module
        self.session.add(unit)

        # Add ref_unit if provided
        if ref_unit:
            ref_unit_obj = model.RefUnit(unit=ref_unit)
            ref_unit_obj.unit = unit
            self.session.add(ref_unit_obj)

        # Add unit_conversion if provided
        if unit_conversion:
            si_exponents = model.SiExponents()
            si_exponents.unit = unit
            self.session.add(si_exponents)

            unit_conversion_obj = model.UnitConversion(
                gradient=unit_conversion.get("gradient", 1.0), offset=unit_conversion.get("offset", 0.0)
            )
            unit_conversion_obj.unit = unit
            self.session.add(unit_conversion_obj)

        return unit

    def add_compu_tab(
        self,
        module: model.Module,
        name: str,
        long_identifier: str,
        conversion_type: str,
        pairs: List[Tuple[float, float]],
        default_numeric: Optional[float] = None,
    ) -> model.CompuTab:
        """Add a numeric COMPU_TAB to a Module.

        Parameters
        ----------
        module : model.Module
            Module owning the table
        name : str
            Table name
        long_identifier : str
            Description
        conversion_type : str
            One of TAB_INTP, TAB_NOINTP
        pairs : List[Tuple[float, float]]
            List of (inVal, outVal) numeric pairs
        default_numeric : Optional[float]
            Optional default numeric display value
        """
        ct = model.CompuTab(
            name=name,
            longIdentifier=long_identifier,
            conversionType=conversion_type,
            numberValuePairs=len(pairs),
        )
        ct.module = module
        self.session.add(ct)
        # add pairs in order
        for idx, (inp, outp) in enumerate(pairs):
            p = model.CompuTabPair()
            p.position = idx
            p.inVal = float(inp)
            p.outVal = float(outp)
            p.parent = ct
            self.session.add(p)
        if default_numeric is not None:
            d = model.DefaultValueNumeric(display_value=float(default_numeric))
            d.compu_tab = ct
            self.session.add(d)
        return ct

    def add_compu_vtab(
        self,
        module: model.Module,
        name: str,
        long_identifier: str,
        conversion_type: str,
        pairs: List[Tuple[float, str]],
        default_value: Optional[str] = None,
    ) -> model.CompuVtab:
        """Add a COMPU_VTAB (verbal table) to a Module."""
        vt = model.CompuVtab(
            name=name,
            longIdentifier=long_identifier,
            conversionType=conversion_type,
            numberValuePairs=len(pairs),
        )
        vt.module = module
        self.session.add(vt)
        for idx, (inp, outp) in enumerate(pairs):
            p = model.CompuVtabPair()
            p.position = idx
            p.inVal = float(inp)
            p.outVal = str(outp)
            p.parent = vt
            self.session.add(p)
        if default_value is not None:
            dv = model.DefaultValue(value=str(default_value))
            dv.compu_vtab = vt
            self.session.add(dv)
        return vt

    def add_compu_vtab_range(
        self,
        module: model.Module,
        name: str,
        long_identifier: str,
        triples: List[Tuple[float, float, str]],
        default_value: Optional[str] = None,
    ) -> model.CompuVtabRange:
        """Add a COMPU_VTAB_RANGE (verbal ranges) to a Module."""
        vr = model.CompuVtabRange(
            name=name,
            longIdentifier=long_identifier,
            numberValueTriples=len(triples),
        )
        vr.module = module
        self.session.add(vr)
        for idx, (vmin, vmax, outp) in enumerate(triples):
            t = model.CompuVtabRangeTriple()
            t.position = idx
            t.inValMin = float(vmin)
            t.inValMax = float(vmax)
            t.outVal = str(outp)
            t.parent = vr
            self.session.add(t)
        if default_value is not None:
            dv = model.DefaultValue(value=str(default_value))
            dv.compu_vtab_range = vr
            self.session.add(dv)
        return vr

    def add_blob(
        self,
        module: model.Module,
        name: str,
        long_identifier: str,
        address: int,
        length: int,
        calibration_access: Optional[str] = None,
    ) -> model.Blob:
        """Add BLOB to a Module."""
        blob = model.Blob(
            name=name,
            longIdentifier=long_identifier,
            address=address,
            length=length,
        )
        blob.module = module
        self.session.add(blob)
        if calibration_access:
            ca = model.CalibrationAccess(access=calibration_access)
            ca.blob = blob
            self.session.add(ca)
        return blob

    def add_transformer(
        self,
        module: model.Module,
        name: str,
        version: str,
        dllname32: str,
        dllname64: str,
        timeout: int,
        trigger: str,
        reverse: str,
        in_objects: Optional[List[str]] = None,
        out_objects: Optional[List[str]] = None,
    ) -> model.Transformer:
        """Add TRANSFORMER to a Module."""
        tr = model.Transformer(
            name=name,
            version=version,
            dllname32=dllname32,
            dllname64=dllname64,
            timeout=timeout,
            trigger=trigger,
            reverse=reverse,
        )
        tr.module = module
        self.session.add(tr)
        if in_objects is not None:
            ino = model.TransformerInObjects()
            self.session.add(ino)
            for ident in in_objects:
                ino._identifier.append(model.TransformerInObjectsIdentifiers(ident))
            tr.transformer_in_objects = ino
        if out_objects is not None:
            outo = model.TransformerOutObjects()
            self.session.add(outo)
            for ident in out_objects:
                outo._identifier.append(model.TransformerOutObjectsIdentifiers(ident))
            tr.transformer_out_objects = outo
        return tr

    def add_frame(
        self,
        module: model.Module,
        name: str,
        long_identifier: str,
        scaling_unit: int,
        rate: int,
        measurements: Optional[List[str]] = None,
        if_data_texts: Optional[List[str]] = None,
    ) -> model.Frame:
        """Add FRAME to a Module."""
        fr = model.Frame(
            name=name,
            longIdentifier=long_identifier,
            scalingUnit=scaling_unit,
            rate=rate,
        )
        fr.module = module
        self.session.add(fr)
        if measurements:
            fm = model.FrameMeasurement()
            self.session.add(fm)
            for m in measurements:
                fm._identifier.append(model.FrameMeasurementIdentifiers(m))
            fr.frame_measurement = fm
        if if_data_texts:
            for txt in if_data_texts:
                idata = model.IfData(content=txt)
                idata.frame = fr
                self.session.add(idata)
        return fr

    def add_typedef_structure(
        self,
        module: model.Module,
        name: str,
        long_identifier: str,
        size: int,
    ) -> model.TypedefStructure:
        """Add TYPEDEF_STRUCTURE to a Module."""
        ts = model.TypedefStructure(
            name=name,
            longIdentifier=long_identifier,
            size=size,
        )
        ts.module = module
        self.session.add(ts)
        return ts

    def add_structure_component(
        self,
        typedef_structure: model.TypedefStructure,
        name: str,
        type_ref: str,
        offset: int,
    ) -> model.StructureComponent:
        """Add STRUCTURE_COMPONENT to a TypedefStructure."""
        sc = model.StructureComponent(
            name=name,
            type_ref=type_ref,
            offset=offset,
        )
        sc.typedef_structure = typedef_structure
        self.session.add(sc)
        return sc

    def add_typedef_measurement(
        self,
        module: model.Module,
        name: str,
        long_identifier: str,
        datatype: str,
        conversion: str,
        resolution: int,
        accuracy: float,
        lower_limit: float,
        upper_limit: float,
    ) -> model.TypedefMeasurement:
        """Add TYPEDEF_MEASUREMENT to a Module."""
        tm = model.TypedefMeasurement(
            name=name,
            longIdentifier=long_identifier,
            datatype=datatype,
            conversion=conversion,
            resolution=resolution,
            accuracy=accuracy,
            lowerLimit=lower_limit,
            upperLimit=upper_limit,
        )
        tm.module = module
        self.session.add(tm)
        return tm

    def add_typedef_characteristic(
        self,
        module: model.Module,
        name: str,
        long_identifier: str,
        char_type: str,
        deposit: str,
        max_diff: float,
        conversion: str,
        lower_limit: float,
        upper_limit: float,
    ) -> model.TypedefCharacteristic:
        """Add TYPEDEF_CHARACTERISTIC to a Module."""
        tc = model.TypedefCharacteristic(
            name=name,
            longIdentifier=long_identifier,
            type=char_type,
            deposit=deposit,
            maxDiff=max_diff,
            conversion=conversion,
            lowerLimit=lower_limit,
            upperLimit=upper_limit,
        )
        tc.module = module
        self.session.add(tc)
        return tc

    def add_instance(
        self,
        module: model.Module,
        name: str,
        long_identifier: str,
        type_name: str,
        address: int,
    ) -> model.Instance:
        """Add INSTANCE to a Module."""
        inst = model.Instance(
            name=name,
            longIdentifier=long_identifier,
            typeName=type_name,
            address=address,
        )
        inst.module = module
        self.session.add(inst)
        return inst

    def add_mod_common(
        self,
        module: model.Module,
        comment: Optional[str] = None,
        byte_order: Optional[str] = None,
        data_size: Optional[int] = None,
        deposit: Optional[str] = None,
        s_rec_layout: Optional[str] = None,
    ) -> model.ModCommon:
        """Add ModCommon to a Module.

        Parameters
        ----------
        module : model.Module
            The Module to add ModCommon to
        comment : Optional[str], optional
            The comment, by default None
        byte_order : Optional[str], optional
            The byte order (e.g., "MSB_FIRST", "MSB_LAST"), by default None
        data_size : Optional[int], optional
            The data size, by default None
        deposit : Optional[str], optional
            The deposit mode (e.g., "ABSOLUTE", "DIFFERENCE"), by default None
        s_rec_layout : Optional[str], optional
            The standard record layout, by default None

        Returns
        -------
        model.ModCommon
            The newly created ModCommon instance
        """
        mod_common = model.ModCommon(comment=comment)
        mod_common.module = module
        self.session.add(mod_common)

        # Add byte order if provided
        if byte_order:
            byte_order_obj = model.ByteOrder(byteOrder=byte_order)
            byte_order_obj.mod_common = mod_common
            self.session.add(byte_order_obj)

        # Add data size if provided
        if data_size:
            data_size_obj = model.DataSize(size=data_size)
            data_size_obj.mod_common = mod_common
            self.session.add(data_size_obj)

        # Add deposit if provided
        if deposit:
            deposit_obj = model.Deposit(mode=deposit)
            deposit_obj.mod_common = mod_common
            self.session.add(deposit_obj)

        # Add s_rec_layout if provided
        if s_rec_layout:
            s_rec_layout_obj = model.SRecLayout(name=s_rec_layout)
            s_rec_layout_obj.mod_common = mod_common
            self.session.add(s_rec_layout_obj)

        return mod_common

    def add_mod_par(
        self,
        module: model.Module,
        comment: Optional[str] = None,
        cpu_type: Optional[str] = None,
        customer: Optional[str] = None,
        customer_no: Optional[str] = None,
        ecu: Optional[str] = None,
        ecu_calibration_offset: Optional[int] = None,
        epk: Optional[str] = None,
        phone_no: Optional[str] = None,
        supplier: Optional[str] = None,
        user: Optional[str] = None,
        version: Optional[str] = None,
    ) -> model.ModPar:
        """Add ModPar to a Module.

        Parameters
        ----------
        module : model.Module
            The Module to add ModPar to
        comment : Optional[str], optional
            The comment, by default None
        cpu_type : Optional[str], optional
            The CPU type, by default None
        customer : Optional[str], optional
            The customer name, by default None
        customer_no : Optional[str], optional
            The customer number, by default None
        ecu : Optional[str], optional
            The ECU name, by default None
        ecu_calibration_offset : Optional[int], optional
            The ECU calibration offset, by default None
        epk : Optional[str], optional
            The EPK, by default None
        phone_no : Optional[str], optional
            The phone number, by default None
        supplier : Optional[str], optional
            The supplier name, by default None
        user : Optional[str], optional
            The user name, by default None
        version : Optional[str], optional
            The version, by default None

        Returns
        -------
        model.ModPar
            The newly created ModPar instance
        """
        mod_par = model.ModPar(comment=comment)
        mod_par.module = module
        self.session.add(mod_par)

        # Add cpu_type if provided
        if cpu_type:
            cpu_type_obj = model.CpuType(cPU=cpu_type)
            cpu_type_obj.mod_par = mod_par
            self.session.add(cpu_type_obj)

        # Add customer if provided
        if customer:
            customer_obj = model.Customer(customer=customer)
            customer_obj.mod_par = mod_par
            self.session.add(customer_obj)

        # Add customer_no if provided
        if customer_no:
            customer_no_obj = model.CustomerNo(number=customer_no)
            customer_no_obj.mod_par = mod_par
            self.session.add(customer_no_obj)

        # Add ecu if provided
        if ecu:
            ecu_obj = model.Ecu(controlUnit=ecu)
            ecu_obj.mod_par = mod_par
            self.session.add(ecu_obj)

        # Add ecu_calibration_offset if provided
        if ecu_calibration_offset is not None:
            ecu_calibration_offset_obj = model.EcuCalibrationOffset(offset=ecu_calibration_offset)
            ecu_calibration_offset_obj.mod_par = mod_par
            self.session.add(ecu_calibration_offset_obj)

        # Add epk if provided
        if epk:
            epk_obj = model.Epk(identifier=epk)
            epk_obj.mod_par = mod_par
            self.session.add(epk_obj)

        # Add phone_no if provided
        if phone_no:
            phone_no_obj = model.PhoneNo(telnum=phone_no)
            phone_no_obj.mod_par = mod_par
            self.session.add(phone_no_obj)

        # Add supplier if provided
        if supplier:
            supplier_obj = model.Supplier(manufacturer=supplier)
            supplier_obj.mod_par = mod_par
            self.session.add(supplier_obj)

        # Add user if provided
        if user:
            user_obj = model.User(userName=user)
            user_obj.mod_par = mod_par
            self.session.add(user_obj)

        # Add version if provided
        if version:
            version_obj = model.Version(versionIdentifier=version)
            version_obj.mod_par = mod_par
            self.session.add(version_obj)

        return mod_par


class CharacteristicCreator(Creator):
    """Creator class for Characteristic entities.

    This class provides methods to create new Characteristic instances and
    their related entities.

    Attributes
    ----------
    session : Any
        SQLAlchemy session object used to interact with the database
    """

    def create_characteristic(
        self,
        name: str,
        long_identifier: str,
        char_type: str,
        address: int,
        deposit: str,
        max_diff: float,
        conversion: str,
        lower_limit: float,
        upper_limit: float,
        module_name: Optional[str] = None,
    ) -> model.Characteristic:
        """Create a new Characteristic instance.

        Parameters
        ----------
        name : str
            Name of the characteristic
        long_identifier : str
            Description of the characteristic
        char_type : str
            Type of the characteristic (e.g., "VALUE", "CURVE", "MAP", "CUBOID")
        address : int
            Address of the characteristic
        deposit : str
            Deposit attribute of the characteristic
        max_diff : float
            Maximum difference of the characteristic
        conversion : str
            Name of the computation method for conversion
        lower_limit : float
            Lower limit of the characteristic
        upper_limit : float
            Upper limit of the characteristic
        module_name : Optional[str], optional
            Name of the module to associate with this Characteristic, by default None

        Returns
        -------
        model.Characteristic
            The newly created Characteristic instance
        """
        # Find the module if module_name is provided
        module = None
        if module_name:
            module = self.session.query(model.Module).filter(model.Module.name == module_name).first()
            if not module:
                raise exceptions.StructuralError(f"Module '{module_name}' not found")

        # Create the Characteristic instance
        characteristic = model.Characteristic(
            name=name,
            longIdentifier=long_identifier,
            type=char_type,
            address=address,
            deposit=deposit,
            maxDiff=max_diff,
            conversion=conversion,
            lowerLimit=lower_limit,
            upperLimit=upper_limit,
        )

        # Associate with module if provided
        if module:
            characteristic.module = module

        # Add to session
        self.session.add(characteristic)

        return characteristic

    def add_axis_descr(
        self,
        characteristic: model.Characteristic,
        attribute: str,
        input_quantity: str,
        conversion: str,
        max_axis_points: int,
        lower_limit: float,
        upper_limit: float,
    ) -> model.AxisDescr:
        """Add AxisDescr to a Characteristic.

        Parameters
        ----------
        characteristic : model.Characteristic
            The Characteristic to add AxisDescr to
        attribute : str
            The attribute (e.g., "STD_AXIS", "FIX_AXIS", "COM_AXIS")
        input_quantity : str
            The input quantity
        conversion : str
            Name of the computation method for conversion
        max_axis_points : int
            Maximum number of axis points
        lower_limit : float
            Lower limit of the axis
        upper_limit : float
            Upper limit of the axis

        Returns
        -------
        model.AxisDescr
            The newly created AxisDescr instance
        """
        axis_descr = model.AxisDescr(
            attribute=attribute,
            inputQuantity=input_quantity,
            conversion=conversion,
            maxAxisPoints=max_axis_points,
            lowerLimit=lower_limit,
            upperLimit=upper_limit,
        )
        axis_descr.characteristic = characteristic
        self.session.add(axis_descr)
        return axis_descr

    def add_bit_mask(self, characteristic: model.Characteristic, mask: int) -> model.BitMask:
        """Add BitMask to a Characteristic.

        Parameters
        ----------
        characteristic : model.Characteristic
            The Characteristic to add BitMask to
        mask : int
            The bit mask

        Returns
        -------
        model.BitMask
            The newly created BitMask instance
        """
        bit_mask = model.BitMask(mask=mask)
        bit_mask.characteristic = characteristic
        self.session.add(bit_mask)
        return bit_mask

    def add_byte_order(self, characteristic: model.Characteristic, byte_order: str) -> model.ByteOrder:
        """Add ByteOrder to a Characteristic.

        Parameters
        ----------
        characteristic : model.Characteristic
            The Characteristic to add ByteOrder to
        byte_order : str
            The byte order (e.g., "MSB_FIRST", "MSB_LAST")

        Returns
        -------
        model.ByteOrder
            The newly created ByteOrder instance
        """
        byte_order_obj = model.ByteOrder(byteOrder=byte_order)
        byte_order_obj.characteristic = characteristic
        self.session.add(byte_order_obj)
        return byte_order_obj

    def add_matrix_dim(
        self, characteristic: model.Characteristic, x_dim: int, y_dim: Optional[int] = None, z_dim: Optional[int] = None
    ) -> model.MatrixDim:
        """Add MatrixDim to a Characteristic.

        Parameters
        ----------
        characteristic : model.Characteristic
            The Characteristic to add MatrixDim to
        x_dim : int
            X dimension
        y_dim : Optional[int], optional
            Y dimension, by default None
        z_dim : Optional[int], optional
            Z dimension, by default None

        Returns
        -------
        model.MatrixDim
            The newly created MatrixDim instance
        """
        matrix_dim = model.MatrixDim(xDim=x_dim, yDim=y_dim if y_dim is not None else 0, zDim=z_dim if z_dim is not None else 0)
        matrix_dim.characteristic = characteristic
        self.session.add(matrix_dim)
        return matrix_dim

    def add_dependent_characteristic(
        self, characteristic: model.Characteristic, formula: str, characteristic_names: List[str]
    ) -> model.DependentCharacteristic:
        """Add DependentCharacteristic to a Characteristic.

        Parameters
        ----------
        characteristic : model.Characteristic
            The Characteristic to add DependentCharacteristic to
        formula : str
            The formula
        characteristic_names : List[str]
            List of characteristic names

        Returns
        -------
        model.DependentCharacteristic
            The newly created DependentCharacteristic instance
        """
        dependent_characteristic = model.DependentCharacteristic(formula=formula)
        dependent_characteristic.characteristic = characteristic
        self.session.add(dependent_characteristic)

        # Add characteristic identifiers
        for char_name in characteristic_names:
            char_id = model.DependentCharacteristicIdentifiers(characteristic=char_name)
            char_id.dependent_characteristic = dependent_characteristic
            self.session.add(char_id)

        return dependent_characteristic

    def add_virtual_characteristic(
        self, characteristic: model.Characteristic, formula: str, characteristic_names: List[str]
    ) -> model.VirtualCharacteristic:
        """Add VirtualCharacteristic to a Characteristic.

        Parameters
        ----------
        characteristic : model.Characteristic
            The Characteristic to add VirtualCharacteristic to
        formula : str
            The formula
        characteristic_names : List[str]
            List of characteristic names

        Returns
        -------
        model.VirtualCharacteristic
            The newly created VirtualCharacteristic instance
        """
        virtual_characteristic = model.VirtualCharacteristic(formula=formula)
        virtual_characteristic.characteristic = characteristic
        self.session.add(virtual_characteristic)

        # Add characteristic identifiers
        for char_name in characteristic_names:
            char_id = model.VirtualCharacteristicIdentifiers(characteristic=char_name)
            char_id.virtual_characteristic = virtual_characteristic
            self.session.add(char_id)

        return virtual_characteristic

    def add_format(self, characteristic: model.Characteristic, format_string: str) -> model.Format:
        """Add Format to a Characteristic.

        Parameters
        ----------
        characteristic : model.Characteristic
            The Characteristic to add Format to
        format_string : str
            The format string

        Returns
        -------
        model.Format
            The newly created Format instance
        """
        format_obj = model.Format(formatString=format_string)
        format_obj.characteristic = characteristic
        self.session.add(format_obj)
        return format_obj

    def add_phys_unit(self, characteristic: model.Characteristic, unit: str) -> model.PhysUnit:
        """Add PhysUnit to a Characteristic.

        Parameters
        ----------
        characteristic : model.Characteristic
            The Characteristic to add PhysUnit to
        unit : str
            The physical unit

        Returns
        -------
        model.PhysUnit
            The newly created PhysUnit instance
        """
        phys_unit = model.PhysUnit(unit=unit)
        phys_unit.characteristic = characteristic
        self.session.add(phys_unit)
        return phys_unit

    def add_symbol_link(
        self, characteristic: model.Characteristic, symbol_name: str, offset: Optional[int] = None
    ) -> model.SymbolLink:
        """Add SymbolLink to a Characteristic.

        Parameters
        ----------
        characteristic : model.Characteristic
            The Characteristic to add SymbolLink to
        symbol_name : str
            The symbol name
        offset : Optional[int], optional
            The offset, by default None

        Returns
        -------
        model.SymbolLink
            The newly created SymbolLink instance
        """
        symbol_link = model.SymbolLink(symbolName=symbol_name, offset=offset if offset is not None else 0)
        symbol_link.characteristic = characteristic
        self.session.add(symbol_link)
        return symbol_link


class FunctionCreator(Creator):
    """Creator class for Function entities.

    This class provides methods to create new Function instances and
    their related entities (DefCharacteristic, FunctionVersion, etc.).

    Attributes
    ----------
    session : Any
        SQLAlchemy session object used to interact with the database
    """

    def create_function(self, name: str, long_identifier: str, module_name: Optional[str] = None) -> model.Function:
        """Create a new Function instance.

        Parameters
        ----------
        name : str
            Name of the function
        long_identifier : str
            Description of the function
        module_name : Optional[str], optional
            Name of the module to associate with this Function, by default None

        Returns
        -------
        model.Function
            The newly created Function instance
        """
        # Find the module if module_name is provided
        module = None
        if module_name:
            module = self.session.query(model.Module).filter(model.Module.name == module_name).first()
            if not module:
                raise exceptions.StructuralError(f"Module '{module_name}' not found")

        # Create the Function instance
        function = model.Function(name=name, longIdentifier=long_identifier)

        # Associate with module if provided
        if module:
            function.module = module

        # Add to session
        self.session.add(function)

        return function

    def add_def_characteristic(self, function: model.Function, characteristic_names: List[str]) -> model.DefCharacteristic:
        """Add DefCharacteristic to a Function.

        Parameters
        ----------
        function : model.Function
            The Function to add DefCharacteristic to
        characteristic_names : List[str]
            List of characteristic names

        Returns
        -------
        model.DefCharacteristic
            The newly created DefCharacteristic instance
        """
        def_characteristic = model.DefCharacteristic()
        def_characteristic.function = function
        self.session.add(def_characteristic)

        # Add characteristic identifiers
        for char_name in characteristic_names:
            char_id = model.DefCharacteristicIdentifiers(identifier=char_name)
            char_id.parent = def_characteristic
            self.session.add(char_id)

        return def_characteristic

    def add_function_version(self, function: model.Function, version_identifier: str) -> model.FunctionVersion:
        """Add FunctionVersion to a Function.

        Parameters
        ----------
        function : model.Function
            The Function to add FunctionVersion to
        version_identifier : str
            The version identifier

        Returns
        -------
        model.FunctionVersion
            The newly created FunctionVersion instance
        """
        function_version = model.FunctionVersion(versionIdentifier=version_identifier)
        function_version.function = function
        self.session.add(function_version)
        return function_version

    def add_in_measurement(self, function: model.Function, measurement_names: List[str]) -> model.InMeasurement:
        """Add InMeasurement to a Function.

        Parameters
        ----------
        function : model.Function
            The Function to add InMeasurement to
        measurement_names : List[str]
            List of measurement names

        Returns
        -------
        model.InMeasurement
            The newly created InMeasurement instance
        """
        in_measurement = model.InMeasurement()
        in_measurement.function = function
        self.session.add(in_measurement)

        # Add measurement identifiers
        for meas_name in measurement_names:
            meas_id = model.InMeasurementIdentifiers(identifier=meas_name)
            meas_id.parent = in_measurement
            self.session.add(meas_id)

        return in_measurement

    def add_loc_measurement(self, function: model.Function, measurement_names: List[str]) -> model.LocMeasurement:
        """Add LocMeasurement to a Function.

        Parameters
        ----------
        function : model.Function
            The Function to add LocMeasurement to
        measurement_names : List[str]
            List of measurement names

        Returns
        -------
        model.LocMeasurement
            The newly created LocMeasurement instance
        """
        loc_measurement = model.LocMeasurement()
        loc_measurement.function = function
        self.session.add(loc_measurement)

        # Add measurement identifiers
        for meas_name in measurement_names:
            meas_id = model.LocMeasurementIdentifiers(identifier=meas_name)
            meas_id.parent = loc_measurement
            self.session.add(meas_id)

        return loc_measurement

    def add_out_measurement(self, function: model.Function, measurement_names: List[str]) -> model.OutMeasurement:
        """Add OutMeasurement to a Function.

        Parameters
        ----------
        function : model.Function
            The Function to add OutMeasurement to
        measurement_names : List[str]
            List of measurement names

        Returns
        -------
        model.OutMeasurement
            The newly created OutMeasurement instance
        """
        out_measurement = model.OutMeasurement()
        out_measurement.function = function
        self.session.add(out_measurement)

        # Add measurement identifiers
        for meas_name in measurement_names:
            meas_id = model.OutMeasurementIdentifiers(identifier=meas_name)
            meas_id.parent = out_measurement
            self.session.add(meas_id)

        return out_measurement

    def add_sub_function(self, function: model.Function, function_names: List[str]) -> model.SubFunction:
        """Add SubFunction to a Function.

        Parameters
        ----------
        function : model.Function
            The Function to add SubFunction to
        function_names : List[str]
            List of function names

        Returns
        -------
        model.SubFunction
            The newly created SubFunction instance
        """
        sub_function = model.SubFunction()
        sub_function.function = function
        self.session.add(sub_function)

        # Add function identifiers
        for func_name in function_names:
            func_id = model.SubFunctionIdentifiers(identifier=func_name)
            func_id.parent = sub_function
            self.session.add(func_id)

        return sub_function


class GroupCreator(Creator):
    """Creator class for Group entities.

    This class provides methods to create new Group instances and
    their related entities (RefMeasurement, RefCharacteristic, etc.).

    Attributes
    ----------
    session : Any
        SQLAlchemy session object used to interact with the database
    """

    def create_group(self, group_name: str, group_long_identifier: str, module_name: Optional[str] = None) -> model.Group:
        """Create a new Group instance.

        Parameters
        ----------
        group_name : str
            Name of the group
        group_long_identifier : str
            Description of the group
        module_name : Optional[str], optional
            Name of the module to associate with this Group, by default None

        Returns
        -------
        model.Group
            The newly created Group instance
        """
        # Find the module if module_name is provided
        module = None
        if module_name:
            module = self.session.query(model.Module).filter(model.Module.name == module_name).first()
            if not module:
                raise exceptions.StructuralError(f"Module '{module_name}' not found")

        # Create the Group instance
        group = model.Group(groupName=group_name, groupLongIdentifier=group_long_identifier)

        # Associate with module if provided
        if module:
            group.module = module

        # Add to session
        self.session.add(group)

        return group

    def add_ref_measurement(self, group: model.Group, measurement_names: List[str]) -> model.RefMeasurement:
        """Add RefMeasurement to a Group.

        Parameters
        ----------
        group : model.Group
            The Group to add RefMeasurement to
        measurement_names : List[str]
            List of measurement names

        Returns
        -------
        model.RefMeasurement
            The newly created RefMeasurement instance
        """
        ref_measurement = model.RefMeasurement()
        ref_measurement.group = group
        self.session.add(ref_measurement)

        # Add measurement identifiers
        for meas_name in measurement_names:
            meas_id = model.RefMeasurementIdentifiers(identifier=meas_name)
            meas_id.parent = ref_measurement
            self.session.add(meas_id)

        return ref_measurement

    def add_ref_characteristic(self, group: model.Group, characteristic_names: List[str]) -> model.RefCharacteristic:
        """Add RefCharacteristic to a Group.

        Parameters
        ----------
        group : model.Group
            The Group to add RefCharacteristic to
        characteristic_names : List[str]
            List of characteristic names

        Returns
        -------
        model.RefCharacteristic
            The newly created RefCharacteristic instance
        """
        ref_characteristic = model.RefCharacteristic()
        ref_characteristic.group = group
        self.session.add(ref_characteristic)

        # Add characteristic identifiers
        for char_name in characteristic_names:
            char_id = model.RefCharacteristicIdentifiers(identifier=char_name)
            char_id.parent = ref_characteristic
            self.session.add(char_id)

        return ref_characteristic

    def add_function_list(self, group: model.Group, function_names: List[str]) -> model.FunctionList:
        """Add FunctionList to a Group.

        Parameters
        ----------
        group : model.Group
            The Group to add FunctionList to
        function_names : List[str]
            List of function names

        Returns
        -------
        model.FunctionList
            The newly created FunctionList instance
        """
        function_list = model.FunctionList()
        function_list.group = group
        self.session.add(function_list)

        # Add function identifiers
        for func_name in function_names:
            func_id = model.FunctionListIdentifiers(name=func_name)
            func_id.parent = function_list
            self.session.add(func_id)

        return function_list

    def add_sub_group(self, group: model.Group, group_names: List[str]) -> model.SubGroup:
        """Add SubGroup to a Group.

        Parameters
        ----------
        group : model.Group
            The Group to add SubGroup to
        group_names : List[str]
            List of group names

        Returns
        -------
        model.SubGroup
            The newly created SubGroup instance
        """
        sub_group = model.SubGroup()
        sub_group.group = group
        self.session.add(sub_group)

        # Add group identifiers
        for group_name in group_names:
            group_id = model.SubGroupIdentifiers(identifier=group_name)
            group_id.parent = sub_group
            self.session.add(group_id)

        return sub_group

    def add_root(self, group: model.Group) -> model.Root:
        """Add Root to a Group.

        Parameters
        ----------
        group : model.Group
            The Group to add Root to

        Returns
        -------
        model.Root
            The newly created Root instance
        """
        root = model.Root()
        root.group = group
        self.session.add(root)
        return root


class RecordLayoutCreator(Creator):
    """Creator class for RecordLayout entities.

    This class provides methods to create new RecordLayout instances and
    their related entities (AlignmentByte, FncValues, etc.).

    Attributes
    ----------
    session : Any
        SQLAlchemy session object used to interact with the database
    """

    def create_record_layout(self, name: str, module_name: Optional[str] = None) -> model.RecordLayout:
        """Create a new RecordLayout instance.

        Parameters
        ----------
        name : str
            Name of the record layout
        module_name : Optional[str], optional
            Name of the module to associate with this RecordLayout, by default None

        Returns
        -------
        model.RecordLayout
            The newly created RecordLayout instance
        """
        # Find the module if module_name is provided
        module = None
        if module_name:
            module = self.session.query(model.Module).filter(model.Module.name == module_name).first()
            if not module:
                raise exceptions.StructuralError(f"Module '{module_name}' not found")

        # Create the RecordLayout instance
        record_layout = model.RecordLayout(name=name)

        # Associate with module if provided
        if module:
            record_layout.module = module

        # Add to session
        self.session.add(record_layout)

        return record_layout

    def add_alignment_byte(self, record_layout: model.RecordLayout, alignment_border: int) -> model.AlignmentByte:
        """Add AlignmentByte to a RecordLayout.

        Parameters
        ----------
        record_layout : model.RecordLayout
            The RecordLayout to add AlignmentByte to
        alignment_border : int
            The alignment border

        Returns
        -------
        model.AlignmentByte
            The newly created AlignmentByte instance
        """
        alignment_byte = model.AlignmentByte(alignmentBorder=alignment_border)
        alignment_byte.record_layout = record_layout
        self.session.add(alignment_byte)
        return alignment_byte

    def add_alignment_word(self, record_layout: model.RecordLayout, alignment_border: int) -> model.AlignmentWord:
        """Add AlignmentWord to a RecordLayout.

        Parameters
        ----------
        record_layout : model.RecordLayout
            The RecordLayout to add AlignmentWord to
        alignment_border : int
            The alignment border

        Returns
        -------
        model.AlignmentWord
            The newly created AlignmentWord instance
        """
        alignment_word = model.AlignmentWord(alignmentBorder=alignment_border)
        alignment_word.record_layout = record_layout
        self.session.add(alignment_word)
        return alignment_word

    def add_alignment_long(self, record_layout: model.RecordLayout, alignment_border: int) -> model.AlignmentLong:
        """Add AlignmentLong to a RecordLayout.

        Parameters
        ----------
        record_layout : model.RecordLayout
            The RecordLayout to add AlignmentLong to
        alignment_border : int
            The alignment border

        Returns
        -------
        model.AlignmentLong
            The newly created AlignmentLong instance
        """
        alignment_long = model.AlignmentLong(alignmentBorder=alignment_border)
        alignment_long.record_layout = record_layout
        self.session.add(alignment_long)
        return alignment_long

    def add_alignment_float32_ieee(self, record_layout: model.RecordLayout, alignment_border: int) -> model.AlignmentFloat32Ieee:
        """Add AlignmentFloat32Ieee to a RecordLayout.

        Parameters
        ----------
        record_layout : model.RecordLayout
            The RecordLayout to add AlignmentFloat32Ieee to
        alignment_border : int
            The alignment border

        Returns
        -------
        model.AlignmentFloat32Ieee
            The newly created AlignmentFloat32Ieee instance
        """
        alignment_float32_ieee = model.AlignmentFloat32Ieee(alignmentBorder=alignment_border)
        alignment_float32_ieee.record_layout = record_layout
        self.session.add(alignment_float32_ieee)
        return alignment_float32_ieee

    def add_alignment_float64_ieee(self, record_layout: model.RecordLayout, alignment_border: int) -> model.AlignmentFloat64Ieee:
        """Add AlignmentFloat64Ieee to a RecordLayout.

        Parameters
        ----------
        record_layout : model.RecordLayout
            The RecordLayout to add AlignmentFloat64Ieee to
        alignment_border : int
            The alignment border

        Returns
        -------
        model.AlignmentFloat64Ieee
            The newly created AlignmentFloat64Ieee instance
        """
        alignment_float64_ieee = model.AlignmentFloat64Ieee(alignmentBorder=alignment_border)
        alignment_float64_ieee.record_layout = record_layout
        self.session.add(alignment_float64_ieee)
        return alignment_float64_ieee

    def add_fnc_values(
        self,
        record_layout: model.RecordLayout,
        position: int,
        data_type: str,
        index_mode: Optional[str] = None,
        address_type: Optional[str] = None,
    ) -> model.FncValues:
        """Add FncValues to a RecordLayout.

        Parameters
        ----------
        record_layout : model.RecordLayout
            The RecordLayout to add FncValues to
        position : int
            The position
        data_type : str
            The data type
        index_mode : Optional[str], optional
            The index mode, by default None
        address_type : Optional[str], optional
            The address type, by default None

        Returns
        -------
        model.FncValues
            The newly created FncValues instance
        """
        fnc_values = model.FncValues(position=position, datatype=data_type, indexMode=index_mode, addresstype=address_type)
        fnc_values.record_layout = record_layout
        self.session.add(fnc_values)
        return fnc_values

    def add_axis_pts_x(
        self,
        record_layout: model.RecordLayout,
        position: int,
        data_type: str,
        index_incr: Optional[str] = None,
        addressing: Optional[str] = None,
    ) -> model.AxisPtsX:
        """Add AxisPtsX to a RecordLayout.

        Parameters
        ----------
        record_layout : model.RecordLayout
            The RecordLayout to add AxisPtsX to
        position : int
            The position
        data_type : str
            The data type
        index_incr : Optional[str], optional
            The index increment, by default None
        addressing : Optional[str], optional
            The addressing, by default None

        Returns
        -------
        model.AxisPtsX
            The newly created AxisPtsX instance
        """
        axis_pts_x = model.AxisPtsX(position=position, datatype=data_type, indexIncr=index_incr, addressing=addressing)
        axis_pts_x.record_layout = record_layout
        self.session.add(axis_pts_x)
        return axis_pts_x

    def add_no_axis_pts_x(self, record_layout: model.RecordLayout, position: int, data_type: str) -> model.NoAxisPtsX:
        """Add NoAxisPtsX to a RecordLayout.

        Parameters
        ----------
        record_layout : model.RecordLayout
            The RecordLayout to add NoAxisPtsX to
        position : int
            The position
        data_type : str
            The data type

        Returns
        -------
        model.NoAxisPtsX
            The newly created NoAxisPtsX instance
        """
        no_axis_pts_x = model.NoAxisPtsX(position=position, datatype=data_type)
        no_axis_pts_x.record_layout = record_layout
        self.session.add(no_axis_pts_x)
        return no_axis_pts_x

    def add_fix_no_axis_pts_x(self, record_layout: model.RecordLayout, number: int) -> model.FixNoAxisPtsX:
        """Add FixNoAxisPtsX to a RecordLayout.

        Parameters
        ----------
        record_layout : model.RecordLayout
            The RecordLayout to add FixNoAxisPtsX to
        number : int
            The number of axis points

        Returns
        -------
        model.FixNoAxisPtsX
            The newly created FixNoAxisPtsX instance
        """
        fix_no_axis_pts_x = model.FixNoAxisPtsX(numberOfAxisPoints=number)
        fix_no_axis_pts_x.record_layout = record_layout
        self.session.add(fix_no_axis_pts_x)
        return fix_no_axis_pts_x



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import contextlib
import pkgutil
from pprint import pprint

import pytest

from pya2l import parsers

ETAS1 = """
    /begin IF_DATA ASAP1B_Bypass
                /begin KP_BLOB
                    BUFFER_OFFSET    "VSwVlv_r_Bypass.Vector"
                    SOURCE_ID        "VSwVlv_r_Bypass.Channel"
                    /* no BIT_OFFSET       */
                    /* no POSSIBLE_SOURCES */
                /end KP_BLOB
    /end IF_DATA
"""

ETAS2 = "/begin IF_DATA ASAP1B_CCP KP_BLOB 0x00 0xD0001240 0x2 RASTER 1 /end IF_DATA"
ETAS3 = "/begin IF_DATA ETK KP_BLOB 0xD0001240 INTERN 0x2 RASTER 2 /end IF_DATA"

SEGMENT = """
/begin IF_DATA XCP
  /begin SEGMENT
    0x0
    0x2
    0x0
    0x0
    0x0
    /begin CHECKSUM
      XCP_ADD_12
      MAX_BLOCK_SIZE 0x0
      EXTERNAL_FUNCTION ""
    /end CHECKSUM
    /begin PAGE
      0x1
      ECU_ACCESS_DONT_CARE
      XCP_READ_ACCESS_DONT_CARE
      XCP_WRITE_ACCESS_DONT_CARE
      INIT_SEGMENT 0x0
    /end PAGE
    /begin PAGE
      0x0
      ECU_ACCESS_DONT_CARE
      XCP_READ_ACCESS_DONT_CARE
      XCP_WRITE_ACCESS_DONT_CARE
      INIT_SEGMENT 0x0
    /end PAGE
    /begin ADDRESS_MAPPING
      0xE1058 0xE1058 0xE9D
    /end ADDRESS_MAPPING
  /end SEGMENT
/end IF_DATA
"""

BASIC = """
    /begin IF_DATA XCP
      /begin PROTOCOL_LAYER
        0x100 0x20 0x20 0x20 0x20 0x20 0x20 0x20 0x8 0x8 BYTE_ORDER_MSB_LAST ADDRESS_GRANULARITY_BYTE
        OPTIONAL_CMD SET_REQUEST
        OPTIONAL_CMD GET_SEED
        OPTIONAL_CMD UNLOCK
        OPTIONAL_CMD SET_MTA
        OPTIONAL_CMD SHORT_DOWNLOAD
        OPTIONAL_CMD PROGRAM_START
        OPTIONAL_CMD PROGRAM_CLEAR
        OPTIONAL_CMD PROGRAM
        OPTIONAL_CMD PROGRAM_RESET
        OPTIONAL_CMD GET_PGM_PROCESSOR_INFO
        OPTIONAL_CMD GET_SECTOR_INFO
        OPTIONAL_CMD PROGRAM_PREPARE
        OPTIONAL_CMD PROGRAM_NEXT
        OPTIONAL_CMD PROGRAM_MAX
        OPTIONAL_CMD PROGRAM_VERIFY
        OPTIONAL_CMD TRANSPORT_LAYER_CMD
        OPTIONAL_CMD GET_ID
        OPTIONAL_CMD UPLOAD
        OPTIONAL_CMD SHORT_UPLOAD
        OPTIONAL_CMD BUILD_CHECKSUM
        OPTIONAL_CMD DOWNLOAD_NEXT
        OPTIONAL_CMD SET_CAL_PAGE
        OPTIONAL_CMD GET_CAL_PAGE
        OPTIONAL_CMD CLEAR_DAQ_LIST
        OPTIONAL_CMD SET_DAQ_PTR
        OPTIONAL_CMD WRITE_DAQ
        OPTIONAL_CMD SET_DAQ_LIST_MODE
        OPTIONAL_CMD GET_DAQ_LIST_MODE
        OPTIONAL_CMD START_STOP_DAQ_LIST
        OPTIONAL_CMD FREE_DAQ
        OPTIONAL_CMD ALLOC_DAQ
        OPTIONAL_CMD ALLOC_ODT
        OPTIONAL_CMD ALLOC_ODT_ENTRY
        OPTIONAL_CMD START_STOP_SYNCH
        OPTIONAL_CMD DOWNLOAD
        OPTIONAL_CMD MODIFY_BITS
        OPTIONAL_CMD GET_PAG_PROCESSOR_INFO
        OPTIONAL_CMD GET_SEGMENT_INFO
        OPTIONAL_CMD GET_PAGE_INFO
        OPTIONAL_CMD SET_SEGMENT_MODE
        OPTIONAL_CMD GET_SEGMENT_MODE
        OPTIONAL_CMD COPY_CAL_PAGE
        OPTIONAL_CMD GET_DAQ_CLOCK
        OPTIONAL_CMD READ_DAQ
        OPTIONAL_CMD GET_DAQ_LIST_MODE
        OPTIONAL_CMD DOWNLOAD_MAX
        COMMUNICATION_MODE_SUPPORTED
          BLOCK
            SLAVE
            MASTER 0x14 0x5
      /end PROTOCOL_LAYER
      /begin DAQ
        STATIC 0x3 0x2 0x0 OPTIMISATION_TYPE_DEFAULT ADDRESS_EXTENSION_DAQ IDENTIFICATION_FIELD_TYPE_ABSOLUTE
        GRANULARITY_ODT_ENTRY_SIZE_DAQ_BYTE 0x4 OVERLOAD_INDICATION_EVENT
        PRESCALER_SUPPORTED
        /begin TIMESTAMP_SUPPORTED
          0x1 SIZE_DWORD UNIT_1US
        /end TIMESTAMP_SUPPORTED
        /begin DAQ_LIST
          0x0
          DAQ_LIST_TYPE DAQ
          MAX_ODT 0x5
          MAX_ODT_ENTRIES 0x7
        /end DAQ_LIST
        /begin DAQ_LIST
          0x1
          DAQ_LIST_TYPE DAQ
          MAX_ODT 0x4
          MAX_ODT_ENTRIES 0x7
        /end DAQ_LIST
        /begin DAQ_LIST
          0x2
          DAQ_LIST_TYPE DAQ
          MAX_ODT 0x3
          MAX_ODT_ENTRIES 0x7
        /end DAQ_LIST
        /begin EVENT
          "5ms" "5ms" 0x0 DAQ 0xFF 0x5 0x6 0xFF
        /end EVENT
        /begin EVENT
          "extEvent" "extEvent" 0x1 DAQ 0xFF 0x1 0x9 0xFF
        /end EVENT
      /end DAQ
      /begin XCP_ON_CAN
        0x100
        CAN_ID_BROADCAST 0x52
        CAN_ID_MASTER 0x51
        CAN_ID_SLAVE 0x50
        BAUDRATE 0x7A120
      /end XCP_ON_CAN
    /end IF_DATA
"""


def load_file(name):
    parser = parsers.aml()
    data = str(pkgutil.get_data("pya2l", "examples/{}".format(name)), "ascii")
    result = parser.parseFromString(data)
    return result.listener_result


@pytest.fixture
def vector_aml():
    return load_file("vector.aml")


@pytest.fixture
def etas_aml():
    return load_file("etas.aml")


def test_if_data_segment(vector_aml):
    ip = parsers.if_data(vector_aml)
    res = ip.parse(SEGMENT)
    assert res == {
        "tag": "IF_DATA",
        "value": [
            {
                "SEGMENT": [
                    0,
                    2,
                    0,
                    0,
                    0,
                    {
                        "CHECKSUM": ["XCP_ADD_12", {"MAX_BLOCK_SIZE": 0, "EXTERNAL_FUNCTION": ""}],
                        "PAGE": [
                            [
                                1,
                                "ECU_ACCESS_DONT_CARE",
                                "XCP_READ_ACCESS_DONT_CARE",
                                "XCP_WRITE_ACCESS_DONT_CARE",
                                {"INIT_SEGMENT": 0},
                            ],
                            [
                                0,
                                "ECU_ACCESS_DONT_CARE",
                                "XCP_READ_ACCESS_DONT_CARE",
                                "XCP_WRITE_ACCESS_DONT_CARE",
                                {"INIT_SEGMENT": 0},
                            ],
                        ],
                        "ADDRESS_MAPPING": [[921688, 921688, 3741]],
                    },
                ]
            },
            {},
        ],
    }


def test_basic(vector_aml):
    ip = parsers.if_data(vector_aml)
    res = ip.parse(BASIC)
    assert res == {
        "tag": "IF_DATA",
        "value": [
            {
                "PROTOCOL_LAYER": [
                    256,
                    32,
                    32,
                    32,
                    32,
                    32,
                    32,
                    32,
                    8,
                    8,
                    "BYTE_ORDER_MSB_LAST",
                    "ADDRESS_GRANULARITY_BYTE",
                    {
                        "OPTIONAL_CMD": [
                            "SET_REQUEST",
                            "GET_SEED",
                            "UNLOCK",
                            "SET_MTA",
                            "SHORT_DOWNLOAD",
                            "PROGRAM_START",
                            "PROGRAM_CLEAR",
                            "PROGRAM",
                            "PROGRAM_RESET",
                            "GET_PGM_PROCESSOR_INFO",
                            "GET_SECTOR_INFO",
                            "PROGRAM_PREPARE",
                            "PROGRAM_NEXT",
                            "PROGRAM_MAX",
                            "PROGRAM_VERIFY",
                            "TRANSPORT_LAYER_CMD",
                            "GET_ID",
                            "UPLOAD",
                            "SHORT_UPLOAD",
                            "BUILD_CHECKSUM",
                            "DOWNLOAD_NEXT",
                            "SET_CAL_PAGE",
                            "GET_CAL_PAGE",
                            "CLEAR_DAQ_LIST",
                            "SET_DAQ_PTR",
                            "WRITE_DAQ",
                            "SET_DAQ_LIST_MODE",
                            "GET_DAQ_LIST_MODE",
                            "START_STOP_DAQ_LIST",
                            "FREE_DAQ",
                            "ALLOC_DAQ",
                            "ALLOC_ODT",
                            "ALLOC_ODT_ENTRY",
                            "START_STOP_SYNCH",
                            "DOWNLOAD",
                            "MODIFY_BITS",
                            "GET_PAG_PROCESSOR_INFO",
                            "GET_SEGMENT_INFO",
                            "GET_PAGE_INFO",
                            "SET_SEGMENT_MODE",
                            "GET_SEGMENT_MODE",
                            "COPY_CAL_PAGE",
                            "GET_DAQ_CLOCK",
                            "READ_DAQ",
                            "GET_DAQ_LIST_MODE",
                            "DOWNLOAD_MAX",
                        ],
                        "COMMUNICATION_MODE_SUPPORTED": {"SLAVE": None, "MASTER": [20, 5]},
                    },
                ],
                "DAQ": [
                    "STATIC",
                    3,
                    2,
                    0,
                    "OPTIMISATION_TYPE_DEFAULT",
                    "ADDRESS_EXTENSION_DAQ",
                    "IDENTIFICATION_FIELD_TYPE_ABSOLUTE",
                    "GRANULARITY_ODT_ENTRY_SIZE_DAQ_BYTE",
                    4,
                    "OVERLOAD_INDICATION_EVENT",
                    {
                        "PRESCALER_SUPPORTED": None,
                        "TIMESTAMP_SUPPORTED": [1, "SIZE_DWORD", "UNIT_1US", {}],
                        "DAQ_LIST": [
                            [0, {"DAQ_LIST_TYPE": "DAQ", "MAX_ODT": 5, "MAX_ODT_ENTRIES": 7}],
                            [1, {"DAQ_LIST_TYPE": "DAQ", "MAX_ODT": 4, "MAX_ODT_ENTRIES": 7}],
                            [2, {"DAQ_LIST_TYPE": "DAQ", "MAX_ODT": 3, "MAX_ODT_ENTRIES": 7}],
                        ],
                        "EVENT": [["5ms", "5ms", 0, "DAQ", 255, 5, 6, 255], ["extEvent", "extEvent", 1, "DAQ", 255, 1, 9, 255]],
                    },
                ],
            },
            {"XCP_ON_CAN": [[256, {"CAN_ID_BROADCAST": 82, "CAN_ID_MASTER": 81, "CAN_ID_SLAVE": 80, "BAUDRATE": 500000}], {}]},
        ],
    }


def test_etas1(etas_aml):
    ip = parsers.if_data(etas_aml)
    res = ip.parse(ETAS1)
    assert res == {
        "tag": "IF_DATA",
        "value": {"KP_BLOB": [{"BUFFER_OFFSET": "VSwVlv_r_Bypass.Vector", "SOURCE_ID": "VSwVlv_r_Bypass.Channel"}]},
    }


def test_etas2(etas_aml):
    ip = parsers.if_data(etas_aml)
    res = ip.parse(ETAS2)
    assert res == {"tag": "IF_DATA", "value": {"KP_BLOB": [0, 3489665600, 2, {"RASTER": [1]}]}}


def test_etas3(etas_aml):
    ip = parsers.if_data(etas_aml)
    res = ip.parse(ETAS3)
    assert res == {"tag": "IF_DATA", "value": {"KP_BLOB": [3489665600, "INTERN", 2, {"RASTER": [2]}]}}
